/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Properties;

import com.backup.zipsnap.ng.job.util.UserIO;

/*
 * Class for handling properties
 */
public class Settings {

	private String propertiesFile = null;
	public Properties properties = new Properties();
	
	/* constants */
	public int defaultExitCode = 1;
	public int Kb = 1024;
	public int Mb = 1024*Kb;
	
	/* GUI related settings */
	public boolean guiDebug = false;
	public boolean confirmExit = true;
	public boolean restoreOpenedTabs = true;
	public ArrayList<String> openedTabs = new ArrayList<String>();
	public String lastOpenedPath = null;
	public int openedTabsSelectedIndex = 0;
	public int fileTreeMaxItems = 10000;
	public int textAreaMaxItems = 10000;
	
	/* CLI related settings */
	public int timeTolerance = 10;
	public boolean pauseOnWarning = false;
	public int progressBarStep = 10;
	public String overwriteExisting = "prompt";
	public String deleteExisting = "prompt";
	public int debug = 0;

	public boolean purgeOldCatalogFiles = true; 
	public boolean purgeAfterBackup = true; 
	public boolean logDebugAlways = true; 
	public boolean logFlushAlways = true; 
	public boolean verifyAfterRestore = true; 
	public boolean listFilesOnSummary = true;
	public int summaryMaxFiles = 0;
	public boolean deleteAfterRestore = true;
	public boolean hideProgressDuringBatch = true;
	public boolean catalogDeepCheck = true;
	public boolean printWarningsSummary = true;
	public boolean infoAfterBackup = true;
	public boolean outputActionWhenBackupEmpty = true;
	public boolean removeInvalidSlicesAfterBackup = true;
	public boolean deepFileAnalysis = false;
	public boolean memoryStats = false;
	public int maxSlicesPerVolume = 25000;
	public int maxVolumeSize = 1024*Mb;

	public int sliceSize = 1*Mb;
	public int bufferSize = 64*Kb;


	/*
	 * Costructor
	 */
	public Settings() {
		if (System.getProperty("propertiesFile") != null) {
			this.propertiesFile = System.getProperty("propertiesFile");
		} 
	}

	/*
	 * Load the custom properties
	 */
	public void load() {
		if (propertiesFile != null) {
			UserIO.print(UserIO.out.VERBOSE, "Loading properties file " + propertiesFile);
			loadFile();
			parse();
			save();
		}
	}

	/*
	 * Load properties from file
	 */
	private void loadFile() {
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(new File(propertiesFile));
		} catch (Exception e) {
			UserIO.reportError("Unable to open the properties file  \""+propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
		}
		try {
			properties.load(fis);
		} catch (Exception e) {
			UserIO.reportError("Unable to load the properties file  \""+ propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
		}
		try {
			fis.close();
		} catch (Exception e) {
			UserIO.reportError("Unable to close the properties file  \""+ propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
		}
	}

	/*
	 * Read and apply the settings
	 */
	public void parse() {
		/* CLI properties */
		if (isValid("debug")) {
			String value = properties.getProperty("debug");
			if (value.equals("no"))
				debug = 0;
			else if (value.equals("yes"))
				debug = 1;
			else if (value.equals("verbose"))
				debug = 2;
			else if (value.equals("all"))
				debug = 3;
			else
				UserIO.reportError("Invalid debug property provided (" + value+ "): valid values are no,yes,verbose,all");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] debug = "+ debug);
		}
		if (isValid("overwriteExisting")) {
			String value = properties.getProperty("overwriteExisting");
			if (!value.equals("yes") && !value.equals("no") && !value.equals("prompt"))
				UserIO.reportError("Invalid overwriteExisting property provided ("+ value + "): valid values are yes,no,prompt");
			else
				overwriteExisting = value;
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] overwriteExisting = "+ overwriteExisting);
		}
		if (isValid("deleteExisting")) {
			String value = properties.getProperty("overwriteExisting");
			if (!value.equals("yes") && !value.equals("no") && !value.equals("prompt"))
				UserIO.reportError("Invalid deleteExisting property provided ("+ value + "): valid values are yes,no,prompt");
			else
				deleteExisting = value;
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] deleteExisting = "+ deleteExisting);
		}
		if (isValid("timeTolerance")) {
			timeTolerance = getInt("timeTolerance");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] timeTolerance = "+ timeTolerance);
		}
		if (isValid("progressBarStep")) {
			progressBarStep = getInt("progressBarStep");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] progressBarStep = "+ progressBarStep);
		}
		if (isValid("pauseOnWarning")) {
			pauseOnWarning = getBoolean("pauseOnWarning");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] pauseOnWarning = "+ pauseOnWarning);
		}
		if (isValid("bufferSize")) {
			bufferSize = getInt("bufferSize") * Kb;
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] bufferSize = "+ bufferSize);
		}
		if (isValid("sliceSize")) {
			sliceSize = getInt("sliceSize") * Mb;
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] sliceSize = "+ sliceSize);
		}
		if (isValid("purgeOldCatalogFiles")) {
			purgeOldCatalogFiles = getBoolean("purgeOldCatalogFiles");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] purgeOldCatalogFiles = "+ purgeOldCatalogFiles);
		}
		if (isValid("purgeAfterBackup")) {
			purgeAfterBackup = getBoolean("purgeAfterBackup");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] purgeAfterBackup = "+ purgeAfterBackup);
		}
		if (isValid("logDebugAlways")) {
			logDebugAlways = getBoolean("logDebugAlways");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] logDebugAlways = "+ logDebugAlways);
		}
		if (isValid("logFlushAlways")) {
			logFlushAlways = getBoolean("logFlushAlways");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] logFlushAlways = "+ logFlushAlways);
		}
		if (isValid("verifyAfterRestore")) {
			verifyAfterRestore = getBoolean("verifyAfterRestore");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] verifyAfterRestore = "+ verifyAfterRestore);
		}
		if (isValid("listFilesOnSummary")) {
			listFilesOnSummary = getBoolean("listFilesOnSummary");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] listFilesOnSummary = "+ listFilesOnSummary);
		}
		if (isValid("deleteAfterRestore")) {
			deleteAfterRestore = getBoolean("deleteAfterRestore");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] deleteAfterRestore = "+ deleteAfterRestore);
		}
		if (isValid("hideProgressDuringBatch")) {
			hideProgressDuringBatch = getBoolean("hideProgressDuringBatch");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] hideProgressDuringBatch = "+ hideProgressDuringBatch);
		}
		if (isValid("catalogDeepCheck")) {
			catalogDeepCheck = getBoolean("catalogDeepCheck");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] catalogDeepCheck = "+ catalogDeepCheck);
		}
		if (isValid("printWarningsSummary")) {
			printWarningsSummary = getBoolean("printWarningsSummary");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] printWarningsSummary = "+ printWarningsSummary);
		}
		if (isValid("infoAfterBackup")) {
			infoAfterBackup = getBoolean("infoAfterBackup");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] infoAfterBackup = "+ infoAfterBackup);
		}
		if (isValid("outputActionWhenBackupEmpty")) {
			outputActionWhenBackupEmpty = getBoolean("outputActionWhenBackupEmpty");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] outputActionWhenBackupEmpty = "+ outputActionWhenBackupEmpty);
		}
		if (isValid("removeInvalidSlicesAfterBackup")) {
			removeInvalidSlicesAfterBackup = getBoolean("removeInvalidSlicesAfterBackup");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] removeInvalidSlicesAfterBackup = "+ removeInvalidSlicesAfterBackup);
		}
		if (isValid("memoryStats")) {
			memoryStats = getBoolean("memoryStats");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] memoryStats = "+ memoryStats);
		}
		if (isValid("summaryMaxFiles")) {
			summaryMaxFiles = getInt("summaryMaxFiles");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] summaryMaxFiles = "+ summaryMaxFiles);
		}
		if (isValid("deepFileAnalysis")) {
			deepFileAnalysis = getBoolean("deepFileAnalysis");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] deepFileAnalysis = "+ deepFileAnalysis);
		}
		if (isValid("maxSlicesPerVolume")) {
			maxSlicesPerVolume = getInt("maxSlicesPerVolume");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] maxSlicesPerVolume = "+ maxSlicesPerVolume);
		}
		if (isValid("maxVolumeSize")) {
			maxVolumeSize = getInt("maxVolumeSize") * Mb;
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] maxVolumeSize = "+ maxVolumeSize);
		}		
		
		/* GUI properties */
		if (isValid("guiDebug")) {
			guiDebug = getBoolean("guiDebug");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] guiDebug = "+ guiDebug);
		}
		if (isValid("restoreOpenedTabs")) {
			restoreOpenedTabs = getBoolean("restoreOpenedTabs");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] restoreOpenedTabs = "+ restoreOpenedTabs);
		}
		if (isValid("openedTabs")) {
			openedTabs.clear();
			final String[] files = properties.getProperty("openedTabs").split("\\;");
			for (String file : files) {
				UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] openedTabs = "+ file);
				openedTabs.add(file);
			}  
		}
		if (isValid("confirmExit")) {
			confirmExit = getBoolean("confirmExit");
			UserIO.print(UserIO.out.VERBOSE,"[PROPERTIES] confirmExit = "+ confirmExit);
		}
		if (isValid("openedTabsSelectedIndex")) {
			openedTabsSelectedIndex = getInt("openedTabsSelectedIndex");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] openedTabsSelectedIndex = "+ openedTabsSelectedIndex);
		}
		if (isValid("fileTreeMaxItems")) {
			fileTreeMaxItems = getInt("fileTreeMaxItems");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] fileTreeMaxItems = "+ fileTreeMaxItems);
		}
		if (isValid("textAreaMaxItems")) {
			textAreaMaxItems = getInt("textAreaMaxItems");
			UserIO.print(UserIO.out.VERBOSE, "[PROPERTIES] textAreaMaxItems = "+ textAreaMaxItems);
		}
		
	}

	/*
	 * Check if a property is valid
	 */
	private boolean isValid(String property) {
		if (properties.getProperty(property) != null && !properties.getProperty(property).equals("")) return true;
		return false;
	}

	/*
	 * Parse int properties
	 */
	private int getInt(String property) {
		int i = 0;
		try {
			i = Integer.parseInt(properties.getProperty(property));
		} catch (Exception e) {
			UserIO.reportError("Invalid " + property + " property provided ("+ properties.getProperty(property)+ "): it must be an integer");
		}
		return i;
	}

	/*
	 * Parse boolean properties
	 */
	private boolean getBoolean(String property) {
		if (properties.getProperty(property).equals("true"))
			return true;
		else if (properties.getProperty(property).equals("false"))
			return false;
		else
			UserIO.reportError("Invalid " + property + " property provided ("+ properties.getProperty(property)+ "): it must be true or false");
		return false;
	}
	
	/*
	 * set a property
	 */
	public void set(String name, String value) {
		properties.setProperty(name, value);
	}
	
	/*
	 * save the settings
	 */
	public void save() {
		/* save GUI settings */
		Main.settings.properties.setProperty("confirmExit", Main.settings.confirmExit+"");
		Main.settings.properties.setProperty("restoreOpenedTabs", Main.settings.restoreOpenedTabs+"");
		Main.settings.properties.setProperty("guiDebug", Main.settings.guiDebug+"");
		Main.settings.properties.setProperty("fileTreeMaxItems", Main.settings.fileTreeMaxItems+"");
		Main.settings.properties.setProperty("textAreaMaxItems", Main.settings.textAreaMaxItems+"");
		
		/* save CLI settings */
		String debugString = "";
		if (Main.settings.debug == 0) debugString = "no";
		if (Main.settings.debug == 1) debugString = "yes";
		if (Main.settings.debug == 2) debugString = "verbose";
		if (Main.settings.debug == 3) debugString = "all";
		Main.settings.properties.setProperty("debug", debugString+"");
		
		Main.settings.properties.setProperty("memoryStats", Main.settings.memoryStats+"");
		Main.settings.properties.setProperty("timeTolerance", Main.settings.timeTolerance+"");
		Main.settings.properties.setProperty("pauseOnWarning", Main.settings.pauseOnWarning+"");
		Main.settings.properties.setProperty("overwriteExisting", Main.settings.overwriteExisting+"");
		Main.settings.properties.setProperty("deleteExisting", Main.settings.deleteExisting+"");
		Main.settings.properties.setProperty("progressBarStep", Main.settings.progressBarStep+"");
		Main.settings.properties.setProperty("bufferSize", Main.settings.bufferSize/Main.settings.Kb+"");
		Main.settings.properties.setProperty("sliceSize", Main.settings.sliceSize/Main.settings.Mb+"");
		Main.settings.properties.setProperty("maxSlicesPerVolume", Main.settings.maxSlicesPerVolume+"");
		Main.settings.properties.setProperty("maxVolumeSize", Main.settings.maxVolumeSize/Main.settings.Mb+"");
		
		/* save advanced settings */		
		Main.settings.properties.setProperty("purgeOldCatalogFiles", Main.settings.purgeOldCatalogFiles+"");
		Main.settings.properties.setProperty("purgeAfterBackup", Main.settings.purgeAfterBackup+"");
		Main.settings.properties.setProperty("logDebugAlways", Main.settings.logDebugAlways+"");
		Main.settings.properties.setProperty("logFlushAlways", Main.settings.logFlushAlways+"");
		Main.settings.properties.setProperty("verifyAfterRestore", Main.settings.verifyAfterRestore+"");
		Main.settings.properties.setProperty("listFilesOnSummary", Main.settings.listFilesOnSummary+"");
		Main.settings.properties.setProperty("summaryMaxFiles", Main.settings.summaryMaxFiles+"");
		Main.settings.properties.setProperty("deleteAfterRestore", Main.settings.deleteAfterRestore+"");
		Main.settings.properties.setProperty("hideProgressDuringBatch", Main.settings.hideProgressDuringBatch+"");
		Main.settings.properties.setProperty("catalogDeepCheck", Main.settings.catalogDeepCheck+"");
		Main.settings.properties.setProperty("printWarningsSummary", Main.settings.printWarningsSummary+"");
		Main.settings.properties.setProperty("infoAfterBackup", Main.settings.infoAfterBackup+"");
		Main.settings.properties.setProperty("outputActionWhenBackupEmpty", Main.settings.outputActionWhenBackupEmpty+"");
		Main.settings.properties.setProperty("removeInvalidSlicesAfterBackup", Main.settings.removeInvalidSlicesAfterBackup+"");
		Main.settings.properties.setProperty("deepFileAnalysis", Main.settings.deepFileAnalysis+"");
		saveFile();
	}
	
	private void saveFile() {
		if (propertiesFile != null) {
			FileOutputStream fos = null;
			try {
				fos = new FileOutputStream(new File(propertiesFile));
			} catch (Exception e) {
				UserIO.reportError("Unable to open for writing the properties file  \""+propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
			}
			try {
				properties.store(fos, null);
			} catch (Exception e) {
				UserIO.reportError("Unable to save the properties file  \""+ propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
			}
			try {
				fos.close();
			} catch (Exception e) {
				UserIO.reportError("Unable to close the properties file  \""+ propertiesFile + "\": "+ UserIO.getExceptionMessage(e));
			}
		}
	}
	
}