/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job;

import java.io.File;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import net.lingala.zip4j.io.ZipOutputStream;
import net.lingala.zip4j.model.FileHeader;
import net.lingala.zip4j.model.ZipParameters;
import net.lingala.zip4j.util.Zip4jConstants;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.JobWrapper;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.unit.SliceUnit;
import com.backup.zipsnap.ng.job.unit.SnapshotUnit;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.ProgressBar;
import com.backup.zipsnap.ng.job.util.UserIO;


public class Backup
{
	/**
	 * Add a snapshot of the current directory to the archive
	 */
	public static void doBackup()
	{
		String encrypted = Job.config.get().password != null ? " (encrypted)" : "";
		UserIO.print(UserIO.out.INFO,"Requested action: BACKUP"+encrypted);
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[INITIALIZING]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"INITIALIZING");

		/* sanity check: configured source directories */
		for (FileUnit directory : Job.config.get().sources) {
			UserIO.print(UserIO.out.INFO,"Source directory: " + directory.nativeName);
			/* check if each source directory is valid */
			FileIO.checkDirectory(directory);

			/* check that archive directory is not a subdirectory of the source directory to avoid a loop */
			if (Job.config.get().archive.name.startsWith(directory.name)) {
				UserIO.reportError("The backup directory " + Job.config.get().archive.nativeName +
						" cannot be a subdirectory of the source directory " + directory.nativeName);
				Job.config.get().sources.remove(directory);
			}
		}

		/* display the snapshot number that is going to be created */
		UserIO.print(UserIO.out.INFO,"Creating snapshot #"+ Job.config.get().snapshot);
		/* display filter, if defined */
		if (Job.config.get().filter != null) UserIO.print(UserIO.out.INFO,"Configured Regex filter: " + Job.config.get().filter.pattern());

		/* copy the backup job into the archive directory */
		final FileIO.FileIOResult result1 = FileIO.copyFile(Job.configFile.get(),new File(Job.config.get().archive.file,FileIO.getCustomFilename("config") + ".txt"));
		if (!result1.success)  {
			UserIO.reportWarning(UserIO.out.INFO,"Unable to create the configuration file from "+Job.configFile.get().getPath()+
					" into the archive directory: "+ result1.errorMessage);
		}
		final FileIO.FileIOResult result2 = FileIO.copyFile(Job.configFile.get(),new File(Job.config.get().archive.file,Job.configFile.get().getName()));
		if (!result2.success)  {
			UserIO.reportWarning(UserIO.out.INFO,"Unable to create a copy of the configuration file from "+Job.configFile.get().getPath()+
					" into the archive directory: "+ result2.errorMessage);
		}

		/***************************************
		 * (1) LIST SOURCE DIRECTORY CONTENTS *
		 ***************************************/

		/* get filtered contents of the source directories */
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[ASSESSING SOURCE DIRECTORIES]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"ASSESSING SOURCE DIRECTORIES");
		Job.timer.get().start("ASSESSING SOURCE DIRECTORIES");
		final List<FileUnit> backupDirContents = new ArrayList<FileUnit>();
		long sourceSize = FileIO.listContents(Job.config.get().sources,backupDirContents);
		Job.memoryUsage("backupDirContents",backupDirContents);
		Job.gui(JobWrapper.msg.TASK_LABEL,"Files in the source directories: "+backupDirContents.size()+" ("+FileIO.getSize(sourceSize)+")");
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("ASSESSING SOURCE DIRECTORIES"));

		/**************************************
		 * (2) DETERMINE DIFFERENTIAL BACKUP *
		 *************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[CALCULATING DIFFERENTIAL BACKUP]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"CALCULATING DIFFERENTIAL BACKUP");
		Job.timer.get().start("CALCULATING DIFFERENTIAL BACKUP");
		/* prepare the structure to store files to be included in the snapshot */
		final List<FileUnit> addFiles = new ArrayList<FileUnit>();
		final List<FileUnit> newFiles = new ArrayList<FileUnit>();
		final List<FileUnit> updateFiles = new ArrayList<FileUnit>();
		final List<FileUnit> copyFiles = new ArrayList<FileUnit>();

		UserIO.print(UserIO.out.INFO,"Matching source directory contents against the catalog...",false);
		long progressSourceSize = 0;
		ProgressBar catalogMatchProgressBar = new ProgressBar();
		/* for each file of the source directories */
		for (FileUnit u : backupDirContents) {
			UserIO.print(UserIO.out.VERBOSE,"Evaluating "+u.nativeName+"...",false);
			boolean isOld = false;
			boolean isNew = false;
			boolean isUpdated = false;
			String stats = null;
			FileUnit match = null;
			/* calculates the checksum that will be used for file comparison */
			if (Main.settings.deepFileAnalysis) u.updateChecksum();
			
			/* determine if the file is old, new or updated */
			/* search for a file/directory in the catalog with the same name */
			final int matchIndex = Collections.binarySearch(Job.catalog.get().files, u);
			if (matchIndex >= 0) {
				/* there is at least an item in the catalog with the same name, look for best match among the revisions */
				match = Job.catalog.get().find(u,matchIndex);
				if (match != null) isOld = true;
				else isUpdated = true; 
			} else isNew = true;

			/* take the appropriate action */
			if (! isOld) {
				/* calculates the checksum if not done yet */
				if (! Main.settings.deepFileAnalysis) u.updateChecksum();
				/* analyse new slices */
				for (int i = 0; i < u.sliceList.size(); i++) {
					SliceUnit slice = u.sliceList.get(i);
					/* if it is not in the catalog, it is a new slice*/
					if (!Job.catalog.get().slices.containsKey(slice.checksum)) {
						slice.isNew = true;
						u.newSlices++;
						u.partialSize += slice.size;
						slice.files.add(u);
						/* add the slice to the list of slices */
						Job.catalog.get().slices.put(slice.checksum,slice);
						UserIO.print(UserIO.out.ALL,"(N:"+slice.checksum+")",false);
					} else UserIO.print(UserIO.out.ALL,"(.:"+slice.checksum+")",false);
				}
				/* if there is at least a new slice, prepare the stats */
				stats = (u.newSlices > 0) ? " ["+u.newSlices+"/"+u.numSlices+"] ("+u.percentageNew()+"%)" : "";
			}
			if (isOld) {
				UserIO.print(UserIO.out.VERBOSE,"already in the catalog");
				/* but track the current snapshot */
				match.snapshotsIndex.add(Job.config.get().snapshot-1);
			}
			if (isNew) {
				if (!u.isDirectory && u.newSlices == 0) {
					/* even if the file is new all the contents is already in the catalog so it should be a copy of an existing file(s) */
					UserIO.print(UserIO.out.VERBOSE,"is a copy of an existing item");
					copyFiles.add(u);
				} else {
					UserIO.print(UserIO.out.VERBOSE,"is a new item"+stats);
					newFiles.add(u);
				}
			}
			if (isUpdated) {
				UserIO.print(UserIO.out.VERBOSE,"is an updated item"+stats);
				updateFiles.add(u);
			}
			progressSourceSize += u.size;
			catalogMatchProgressBar.display(progressSourceSize,sourceSize);
		}
		/* update the addFiles list with all the files to be added to the catalog */
		addFiles.addAll(newFiles);
		addFiles.addAll(updateFiles);
		addFiles.addAll(copyFiles);
		Job.memoryUsage("addFiles",addFiles);
		Job.memoryUsage("Catalog",Job.catalog.get());
		UserIO.print(UserIO.out.INFO,"done");
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("CALCULATING DIFFERENTIAL BACKUP"));

		/**********************************************************
		 * (3) PRINT THE BACKUP SUMMARY                            *
		 **********************************************************/

		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[BACKUP SUMMARY]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"BACKUP SUMMARY");
		long newFilesSize = 0;
		long updateFilesSize = 0;
		long copyFilesSize = 0;
		long addFilesSize = 0;
		int counter = 0;
		UserIO.out out = Main.settings.listFilesOnSummary ? UserIO.out.INFO : UserIO.out.DEBUG;
		/* at least a new item to add */
		if (addFiles.size() > 0) {
			/* new files */
			if (newFiles.size() > 0) {
				counter = 0;
				UserIO.print(UserIO.out.INFO,"The following NEW items will be added to the snapshot: ",false);
				for (FileUnit u : newFiles) {
					if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
						UserIO.print(out,"- ...other "+(newFiles.size()-counter)+" items");	
						break;
					}
					String sizeString = u.size > 0 ? ": "+u.percentageNew()+"% ("+FileIO.getSize(u.partialSize)+"/"+FileIO.getSize(u.size)+")" : "";
					UserIO.print(out,"- "+u.nativeName+sizeString);
					Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following NEW items will be added to the snapshot:",u.nativeName,sizeString});
					newFilesSize += u.partialSize;
					counter++;
				}
				UserIO.print(UserIO.out.INFO,newFiles.size()+ " items ("+FileIO.getSize(newFilesSize)+")");
			}
			/* updated files */
			if (updateFiles.size() > 0) {
				counter = 0;
				UserIO.print(UserIO.out.INFO,"The following UPDATED items will be added to the snapshot: ",false);
				for (FileUnit u : updateFiles) {
					if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
						UserIO.print(out,"- ...other "+(updateFiles.size()-counter)+" items");	
						break;
					}
					String sizeString = u.size > 0 ? ": "+u.percentageNew()+"% ("+FileIO.getSize(u.partialSize)+"/"+FileIO.getSize(u.size)+")" : "";
					UserIO.print(out,"- "+u.nativeName+sizeString);
					Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following UPDATED items will be added to the snapshot:",u.nativeName,sizeString});
					updateFilesSize += u.partialSize;
					counter++;
				}
				UserIO.print(UserIO.out.INFO,updateFiles.size()+ " items ("+FileIO.getSize(updateFilesSize)+")");
			}
			/* copies of existing files */
			if (copyFiles.size() > 0) {
				counter = 0;
				UserIO.print(UserIO.out.INFO,"The following COPIES of existing files will be added to the snapshot: ",false);
				for (FileUnit u : copyFiles) {
					if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
						UserIO.print(out,"- ...other "+(copyFiles.size()-counter)+" items");	
						break;
					}
					UserIO.print(out,"- "+u.nativeName);
					Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following COPIES of existing files will be added to the snapshot:",u.nativeName,""});
					copyFilesSize += u.partialSize;
					counter++;
				}
				UserIO.print(UserIO.out.INFO,copyFiles.size()+ " items ("+FileIO.getSize(copyFilesSize)+")");
			}
		} else {
			/* nothing to be done */
			UserIO.print(UserIO.out.INFO,"No items need to be added to the snapshot");
			Job.gui(JobWrapper.msg.NO_NEW_ITEMS, "");
			/* skip two steps */
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"BACKING UP THE FILES");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
		}
		Job.gui(JobWrapper.msg.SUMMARY_READY, "");
		
		if (addFiles.size() > 0) {
			addFilesSize = newFilesSize + updateFilesSize + copyFilesSize;
			UserIO.print(UserIO.out.INFO,"Total items: "+addFiles.size()+" ("+FileIO.getSize(addFilesSize)+")");
			Job.gui(JobWrapper.msg.TASK_LABEL,"Files to backup: "+addFiles.size()+" ("+FileIO.getSize(addFilesSize)+")");
			
			/* wait for user confirmation */
			if (Job.config.get().interactive) {
				Job.timer.get().pause("JOB");
				/* wait for user input */
				char choice = UserIO.userCharPrompt(UserIO.choice.YES_NO,"After reviewing the files which will be included in the snapshot, do you want to proceed with the backup?");
				if (choice == 'N') return;
				Job.timer.get().resume("JOB");
			}

			/**********************************
			 * (4) ADD ITEMS TO A NEW VOLUME *
			 *********************************/
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[BACKING UP THE FILES]");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"BACKING UP THE FILES");
			Job.timer.get().start("BACKING UP THE FILES");
			UserIO.print(UserIO.out.INFO,"Backup in progress...",false);  

			/* data structure to store items added to the snapshot */
			final List<FileUnit> addedFiles = new ArrayList<FileUnit>();

			/* determine zip file name of the first volume */
			final String zipBaseName = FileIO.getCustomFilename("archive");
			File zipFile = new File(Job.config.get().archive.file, zipBaseName + ".zip");

			/* Open the zip archive */
			ZipOutputStream zos = null;

			/* check where the file will be written */
			FileIO.checkFile(zipFile);
			try { zos = new ZipOutputStream(new FileOutputStream(zipFile)); }
			catch (Exception e) {
				UserIO.reportError("Unable to create ZIP volume " + zipFile.getName() + ": " +
						UserIO.getExceptionMessage(e));
			}
			UserIO.print(UserIO.out.DEBUG,"Adding " + addFiles.size() + " items to the volume " + zipFile.getName());

			/* track the snapshot size */
			long snapshotCompressedSize = 0;   
			long snapshotUncompressedSize = 0;  

			/* volume size, used to open a new volume */
			long volumeCompressedSize = 0;
			int volumeNum = 0;  
			int volumeSlices = 0;

			/* for each file to be added */
			ProgressBar backupProgressBar = new ProgressBar();
			AddNextFile:
				for (FileUnit u : addFiles) {
					/* skip this file since already in another volume but add it to the catalog or this is a directory */
					if (u.isDirectory) {
						addedFiles.add(u);
						UserIO.print(UserIO.out.VERBOSE,"Skipping directory "+u.nativeName);
						continue AddNextFile;
					}
					/* ensure no invalid slices are referenced */
					for (SliceUnit slice : u.sliceList) {
						if (Job.catalog.get().slices.get(slice.checksum).isInvalid) {
							UserIO.reportWarning(UserIO.out.DEBUG,"File "+u.nativeName+" references an invalid slice");
							UserIO.reportWarning(UserIO.out.INFO,"Skipping "+u.nativeName);
							u.setInvalid();
							continue AddNextFile;
						}
					}
					/* check if the data to be saved is greater than the file size (may have changed) */
					if (u.partialSize > u.size) {
						UserIO.reportWarning(UserIO.out.DEBUG,"File "+u.nativeName+" has an inconsistent size (size: "+u.size+" to save: "+u.partialSize+")");
						UserIO.reportWarning(UserIO.out.INFO,"Skipping "+u.nativeName);
						u.setInvalid();
						continue AddNextFile;        
					}
					/* skip if the file has no new content */
					if (u.newSlices == 0) {
						addedFiles.add(u);
						UserIO.print(UserIO.out.VERBOSE,"Skipping file without new content "+u.nativeName);
						continue AddNextFile;
					}

					UserIO.print(UserIO.out.DEBUG,"Compressing " + u.nativeName,false);
					/* open the source file with RandomAccessFile*/
					RandomAccessFile ptr = null;
					try { ptr = new RandomAccessFile(u.file, "r"); }
					catch (Exception e) {
						UserIO.reportWarning(UserIO.out.DEBUG,"Unable to open file "+u.file.getPath()+" for reading: " + UserIO.getExceptionMessage(e));
						UserIO.reportWarning(UserIO.out.INFO,"Skipping "+u.nativeName);
						u.setInvalid();
						continue AddNextFile;
					} 

					/* for each slice of the file */
					UserIO.print(UserIO.out.DEBUG," [",false);
					AddNextSlice:
						for (int i = 0; i < u.sliceList.size(); i++) {
							volumeSlices++;
							/* if the maximum archive size is set and this archive is bigger, close it and open a new one */
							if ((Main.settings.maxVolumeSize > 0 && volumeCompressedSize > Main.settings.maxVolumeSize) ||
									(Main.settings.maxSlicesPerVolume > 0 && volumeSlices >= Main.settings.maxSlicesPerVolume)	){
								/* close the current zip volume first */
								try {
									zos.finish();
									zos.close();
								} catch (Exception e) {
									UserIO.reportError("Unable to close ZIP volume " + zipFile.getName() + " (ZIP volume may not be written successfully): " +
											UserIO.getExceptionMessage(e));
								}
								snapshotCompressedSize = zipFile.length();
								/* open the new zip volume */
								volumeNum++;
								zipFile = new File(Job.config.get().archive.file, zipBaseName + "."+ volumeNum + ".zip");
								UserIO.print(UserIO.out.DEBUG,"Current volume is full ("+FileIO.getSize(volumeCompressedSize)+" - "+volumeSlices+" items), opening a new volume: "+zipFile.getName());
								FileIO.checkFile(zipFile);
								try { 
									zos = new ZipOutputStream(new FileOutputStream(zipFile)); 
								} catch (Exception e) {
									UserIO.reportError("Unable to create ZIP volume " + zipFile.getName() + ": " +UserIO.getExceptionMessage(e));
								}
								volumeCompressedSize = 0;
								volumeSlices = 0;
							}

							/* process the next slice */
							SliceUnit slice = u.sliceList.get(i);
							UserIO.print(UserIO.out.ALL,"("+slice.checksum+")",false);
							/* do nothing if it is already in the catalog (e.g. is not new) */
							if (!slice.isNew) {
								UserIO.print(UserIO.out.DEBUG,".",false);
								continue AddNextSlice;
							}
							UserIO.print(UserIO.out.DEBUG,"N",false);
							/* keep track where the current slice will be stored (snapshot/volume) */
							slice.volume= volumeNum;
							slice.snapshot= Job.config.get().snapshot;

							/* filename in the zip file is the same as the checksum */
							ZipParameters parameters = new ZipParameters();
							parameters.setFileNameInZip(slice.checksum);
							parameters.setCompressionMethod(Zip4jConstants.COMP_DEFLATE);
							parameters.setCompressionLevel(Zip4jConstants.DEFLATE_LEVEL_NORMAL);
							if (Job.config.get().password != null) {
								parameters.setEncryptFiles(true);
								parameters.setEncryptionMethod(Zip4jConstants.ENC_METHOD_AES);
								parameters.setAesKeyStrength(Zip4jConstants.AES_STRENGTH_256);
								parameters.setPassword(Job.config.get().password);
							}

							/* write the file to the zip volume */
							FileHeader fileHeader = null;
							try {
								/* commit Zip Entry to the ZIP volume */
								zos.putNextEntry(u.file,parameters);
								/* if this is a file, store it */
								final byte byteBuffer[] = new byte[Main.settings.bufferSize];
								/* move the pointer to the beginning of the current slice */
								ptr.seek(i*u.sliceSize);
								int sliceSize = 0;
								while (true) {
									/* ensure byteBuffer is fully filled in */
									int read = FileIO.fillBuffer(u,ptr,byteBuffer);
									if (read == -1) {
										UserIO.reportWarning(UserIO.out.INFO,"Error while reading "+u.nativeName);
										u.setInvalid();
										continue AddNextFile;
									}
									/* write the buffer to the archive */
									if (read > 0) zos.write(byteBuffer, 0, read);
									sliceSize += read;
									/* exit if EOF or we read too much */
									if (read == 0 || sliceSize == u.sliceSize) break;              
								}
								/* close the zip stream and prepare for next entry */
								fileHeader = zos.closeEntry();
							}
							catch (Exception e) {
								UserIO.reportWarning(UserIO.out.INFO,"Unable to add slice "+(i+1)+" of " + u.nativeName +": " +
										UserIO.getExceptionMessage(e));
								u.setInvalid();
								continue AddNextFile;
							}
							/* check the integrity of the slice that has just been saved */
							if (fileHeader.getCrc32() != slice.crc) {
								/* file is invalid and will not be saved */
								u.setInvalid();
								/* all the slices are marked as invalid and will be purged from the volume */
								UserIO.print(UserIO.out.DEBUG,"(W)",false);
								UserIO.reportWarning(UserIO.out.DEBUG,"Unable to add file "+u.nativeName+" since changed during the backup");
								UserIO.reportWarning(UserIO.out.INFO,"Skipping "+u.nativeName);
								continue AddNextFile;
							}
							/* track compressed size*/
							volumeCompressedSize += fileHeader.getCompressedSize();        
							u.compressedSize += fileHeader.getCompressedSize();   
							/* Show progress */
							snapshotUncompressedSize += slice.size;
							backupProgressBar.display(snapshotUncompressedSize,addFilesSize);
							continue AddNextSlice;
						} /* end for each slice */
					/* close the source file */
					try { ptr.close(); }
					catch (Exception e) {
						UserIO.reportWarning(UserIO.out.INFO,"Unable to close file "+u.nativeName+" after reading: "+ UserIO.getExceptionMessage(e));
					}
					UserIO.print(UserIO.out.DEBUG,"]",false);
					/* show the compression */
					UserIO.print(UserIO.out.DEBUG," (" + FileIO.compressionRatio(u.partialSize,u.compressedSize) + "%)");
					/* file/directory successfully added to ZIP volume */
					u.volume = volumeNum;
					/* check if the size of the data slices is the same as the original file */
					long calculatedSize = 0;
					for (SliceUnit slice : u.sliceList) calculatedSize += Job.catalog.get().slices.get(slice.checksum).size;
					if (calculatedSize > 0 && u.size != calculatedSize)  {
						UserIO.reportWarning(UserIO.out.INFO,"The file "+u.nativeName+" has an inconsisted size (size: " + u.size+" calculated: "+calculatedSize+")");
						u.setInvalid();
					}
					/* succesfully added the file */
					if (u.isInvalid) {
						UserIO.reportWarning(UserIO.out.INFO,"Skipping invalid file " + u.nativeName);
					} else {
						addedFiles.add(u);
					}
					continue AddNextFile;
				} /* end for each file */

			/* close the zip volume */
			try { 
				zos.finish();
				zos.close();
			} catch (Exception e) {
				UserIO.reportError("Unable to close ZIP volume " + zipFile.getName() + ": " +UserIO.getExceptionMessage(e));
			}
			snapshotCompressedSize += zipFile.length();
			UserIO.print(UserIO.out.INFO,"done");

			/* print summary */
			UserIO.print(UserIO.out.INFO,
					"Total items: " + addedFiles.size() + " (uncompressed " +
							FileIO.getSize(snapshotUncompressedSize) + ", compressed " +
							FileIO.getSize(snapshotCompressedSize)+", ratio "+FileIO.compressionRatio(snapshotUncompressedSize,snapshotCompressedSize)+
					"%)");

			/* set timestamp to ZIP volume */
			final FileIO.FileIOResult setTimeResult = FileIO.setFileTime(zipFile, Job.currentTime.get().getTime());
			if (!setTimeResult.success)
				UserIO.reportWarning(UserIO.out.INFO,"Unable to set last-modified time of ZIP volume \"" + zipFile.getName() + "\".");

			Job.memoryUsage("addedFiles",addedFiles);
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("BACKING UP THE FILES"));

			/******************************************
			 * (5) UPDATE THE CATALOG                 *
			 ******************************************/
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[FINALIZING]");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
			Job.timer.get().start("FINALIZING");
			UserIO.print(UserIO.out.INFO,"Updating the catalog...",false);
			/* create a new snapshot and add it to the catalog */
			boolean isEncrypted = Job.config.get().password != null ? true : false;
			final SnapshotUnit s = new SnapshotUnit(Job.config.get().snapshot,Job.config.get().hostname,Job.timestamp.get(),volumeNum,isEncrypted,false);
			Job.catalog.get().snapshotsIndex.add(Job.config.get().snapshot-1, s);

			/* add new/modified files/directories to the catalog */
			for (FileUnit u : addedFiles) {
				/* add this snapshot index to the newly archived file/directory */
				u.snapshotsIndex.add(Job.config.get().snapshot-1);
				UserIO.print(UserIO.out.ALL,"Adding to the catalog "+u.nativeName);
				/* add newly archived file/directory to catalog */
				Job.catalog.get().files.add(u);
			}
			/* sort files/directories in catalog */
			UserIO.print(UserIO.out.DEBUG,"Sorting the catalog");
			Collections.sort(Job.catalog.get().files);
			UserIO.print(UserIO.out.INFO,"done");

			Job.memoryUsage("Catalog",Job.catalog.get());
			/* remove invalid slices from the ZIP archives */
			if (Main.settings.removeInvalidSlicesAfterBackup) Purge.purgeInvalidSlices();

			/*********************************************
			 * (6) WRITE NEW CATALOG FOR THE ARCHIVE     *
			 *********************************************/
			Job.catalog.get().writeToFile(FileIO.getCustomFilename("catalog"));
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("FINALIZING"));

			/**********************************************
			 * (7) CATALOG INTEGRITY CHECK                *
			 *********************************************/
		} // end addFiles.size() > 0
		/* catalog needs to be reloaded to be sure we are working with the latest version */
		UserIO.print(UserIO.out.INFO,"Reloading the catalog...",false);
		Job.catalog.get().reload();
		UserIO.print(UserIO.out.INFO,"done");

		/*********************************************
		 * (8) RUN A CUSTOM COMMAND                 *
		 *********************************************/
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"RUNNING CUSTOM COMMAND");
		if (Job.config.get().runAfterBackup != null && (addFiles.size() > 0 || (addFiles.size() == 0 && Main.settings.outputActionWhenBackupEmpty))) {
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[RUNNING CUSTOM COMMAND]");
			Job.timer.get().start("RUNNING CUSTOM COMMAND");
			/* build the command to run */
			String command = Job.config.get().runAfterBackup+" "+FileIO.getAbsolutePath(Job.config.get().archive.file)+" "+FileIO.getCustomFilename("");
			UserIO.print(UserIO.out.INFO,"Executing command \""+command+"\"...",false);  
			/* run it */
			try {
				Process process = Runtime.getRuntime().exec(command); 
				process.waitFor(); 
			} catch (Exception e) {
				UserIO.reportWarning(UserIO.out.INFO,"There was an error executing the commmand " + command + ": " +UserIO.getExceptionMessage(e));
			}
			UserIO.print(UserIO.out.INFO,"done");
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("RUNNING CUSTOM COMMAND"));
		}

		/******************************************
		 * (9) PURGE THE ARCHIVE                 *
		 *****************************************/
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"PURGING ARCHIVE");
		if (Main.settings.purgeAfterBackup){
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[PURGING ARCHIVE]");
			Job.timer.get().start("PURGING ARCHIVE");
			UserIO.print(UserIO.out.INFO,"Checking purging settings...",false);
			if (Purge.isPurgeConfigured()) {
				UserIO.print(UserIO.out.INFO,"done");
				/* Invoke purge archive */
				Job.timestamp.set(Job.config.get().timestampFormatter.format(new Date()));
				Purge.doPurge(false);
			} else {
				UserIO.print(UserIO.out.INFO,"no purging options configured");
			}
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("PURGING ARCHIVE"));
		}

		/*******************************************
		 * (10) SHOW ARCHIVE INFO                 *
		 *****************************************/
		if (Main.settings.infoAfterBackup) {
			Job.config.get().filter = null;
			Job.gui(JobWrapper.msg.JOB_INFO,"");
			UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[ARCHIVE SUMMARY]");
			Job.timer.get().start("ARCHIVE SUMMARY");
			Info.archiveSummary(false);
			UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("ARCHIVE SUMMARY"));
		}
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");
	}





}