/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map.Entry;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.Settings;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.UserIO;

/**
 * Class for handling configuration settings
 */
public class Config
{
	/* custom properties */
	public String propertiesFile = "PropertiesCLI";
	public Settings properties;
	public static ThreadLocal<Boolean> initialized = new ThreadLocal<Boolean>();
	
	/* source and destination directories for the backup */
	public FileUnit archive = null;
	public List<FileUnit> sources = new ArrayList<FileUnit>();
	public List<FileUnit> directories = new ArrayList<FileUnit>();

	/* contants */
	public SimpleDateFormat timestampFormatter = new SimpleDateFormat("yyyyMMdd-HHmmss", Locale.ENGLISH);
	public SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyyMMdd", Locale.ENGLISH);
	public SimpleDateFormat dateTimeFormatter = new SimpleDateFormat("dd/MM/yyyy HH:mm", Locale.ENGLISH);


	/* global variables */
	public int snapshot = -1;
	public String jobName = null;
	public Pattern filter = null;
	public String password = null;
	public boolean interactive = true;
	public String hostname = null;



	/* backup global variables */
	public String runAfterBackup = null;
	public int purgeAfterDays = -1;
	public String shadowDevice = null;
	public String shadowVolume = null;

	/* restore global variables */
	public enum restoreType {origin,mergeDirectory,extractInto };
	public restoreType restoreMode;
	public FileUnit extractInto = null;
	public HashMap<String,FileUnit> mergeDirectory = new HashMap<String,FileUnit>();




	/**
	 * Costructor
	 */  
	public Config() {
		timestampFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
		Main.settings.pauseOnWarning = false; 
		initialized.set(new Boolean(true));
	}

	/**
	 * Parse a configuration file
	 */
	public void parse(final File configFile) {
		/* generate a job name string by removing the extension if any from the configuration file */
		jobName = configFile.getName();
		if (jobName.indexOf(".") > 0) jobName = jobName.substring(0, jobName.lastIndexOf("."));
		/* determine the hostname */
		hostname = getHostname();

		/* Open the configuration file */
		BufferedReader br = null;
		try { br = new BufferedReader(new InputStreamReader(new FileInputStream(configFile)));
		} catch (Exception e) {
			UserIO.reportError("Unable to open configuration file \"" + configFile.getPath() + "\": " + UserIO.getExceptionMessage(e));
		} 
		UserIO.print(UserIO.out.DEBUG,"Hostname: "+hostname);

		/* read and parse each line of the file */
		final Pattern configLinePattern = Pattern.compile("\\s*([^\\s]+)\\s*=\\s*(.*)\\s*");
		ParseNextLine:
			for (int lineNum = 1; ; lineNum++) {
				String lineString = null;
				/* read the line */
				try {
					if (!br.ready()) break ParseNextLine;
					lineString = br.readLine().trim();
				} catch (Exception e) {
					UserIO.reportError("Unable to read line " + lineNum + " of the configuration file \"" + configFile.getPath() + "\": " + UserIO.getExceptionMessage(e));
				}
				/* skip this line if empty or if it begins with '#' */
				if (lineString.isEmpty() || lineString.startsWith("#")) continue ParseNextLine;
				/* apply the pattern */
				final Matcher m = configLinePattern.matcher(lineString);
				if (!m.matches()) {
					UserIO.reportError("Unable to read line " + lineNum + " of the configuration file \"" + configFile.getPath() + "\": line is malformed");
				}
				final String name = m.group(1);
				final String value = m.group(2);
				UserIO.print(UserIO.out.VERBOSE,"[CONFIG-LINE] '"+name+"' = '"+value+"'");

				/* parse all the switches */
				string2setting(name,value);
				continue ParseNextLine;
			}

		/* check for invalid combinations or missing settings */
		if (shadowDevice != null) {
			UserIO.print(UserIO.out.DEBUG,"Shadow copy device: "+shadowDevice);
		}
		/* check for inconsisten buffer sizes */
		if (Main.settings.bufferSize > Main.settings.sliceSize) UserIO.reportError("bufferSize ("+Main.settings.bufferSize+") cannot be bigger than sliceSize ("+Main.settings.sliceSize+")");
		/* check if the archive directory exists */
		FileIO.checkDirectory(archive);
		if (Job.command.get().equals("backup")) {
			if (sources.isEmpty()) UserIO.reportError("directoriesToBackup cannot be empty");
			Pattern drivePattern = Pattern.compile("^([A-Za-z]):");
			/* for each source directory */
			for (FileUnit directory : sources) {
				/* check if the directory exists */
				FileIO.checkDirectory(directory);
				/* if shadow copy is enabled */
				if (shadowDevice != null && System.getProperty("os.name").toLowerCase().contains("windows")) {
					Matcher m = drivePattern.matcher(directory.nativeName);
					if(m.find()) {
						String drive = m.group(1);
						if (shadowVolume == null) shadowVolume = drive;
						if (shadowVolume != drive) UserIO.reportError("When shadow copy is enabled all the source directories must be in the same drive. "+
								directory.nativeName+" is not on drive "+shadowVolume+":\\");
					} else UserIO.reportError("Cannot determine the drive letter of "+directory.nativeName);
				}
			}
		} 
		else if (Job.command.get().equals("restore")) {
			if (restoreMode == Config.restoreType.extractInto) FileIO.checkDirectory(extractInto);
			if (restoreMode == Config.restoreType.mergeDirectory) {
				for (FileUnit directory : mergeDirectory.values()) {
					FileIO.checkDirectory(directory); 
				} 
			}
		} 

	}

	/*
	 * read and set a given configuration setting
	 */
	public void string2setting(String what, String value) {
		if (what.equals("directoriesToBackup")) {
			sources.clear();
			if (value.equals("")) return;
			/* split the directories and store them into the sources array */
			final String[] directories = value.split("\\;");
			for (String directory : directories) {
				final FileUnit u = new FileUnit(FileIO.getAbsolutePath(directory),"",true);
				UserIO.print(UserIO.out.VERBOSE,"[CONFIG] source directory = "+u.nativeName);
				sources.add(u);
			}  
		}
		else if (what.equals("backupDirectory")) {
			if (value.equals("")) return;
			archive = new FileUnit(FileIO.getAbsolutePath(value),"",true,false);
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] destination directory = "+archive.nativeName);
		}
		else if (what.equals("interactive")) {
			if (value.equals("yes")) return;
			interactive = false;
			/* overwrite and delete by default */ 
			Main.settings.overwriteExisting = "yes";
			Main.settings.deleteExisting = "yes";
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] interactive = "+interactive);
		}
		else if (what.equals("runAfterBackup")) {
			if (value.equals("")) return;
			runAfterBackup = value;
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] runAfterBackup = "+runAfterBackup);
		}
		else if (what.equals("password")) {
			if (value.equals("")) return;
			password = value;
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] password = "+password);
		}
		else if (what.equals("filter") ) {
			if (value.equals("")) return;
			try {
				filter = Pattern.compile(value);
			} catch (Exception e) {
				UserIO.reportError("Unable to compile the specified regex filter string \"" +
						value+ "\": " + UserIO.getExceptionMessage(e));
			}
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] filter = "+filter.pattern());
		}
		else if (what.equals("purgeAfterDays")) {
			if (value.equals("")) return;
			try {
				purgeAfterDays = Integer.parseInt(value);
			} catch (Exception e) {
				purgeAfterDays = -1;
			}
			if (purgeAfterDays < 0) {
				UserIO.reportError("Invalid purgeAfterDays parameter \"" + value +"\" specified: must be a nonnegative integer");
			}
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] purgeAfterDays = "+purgeAfterDays);
		}
		else if (what.equals("restoreInto")) {
			mergeDirectory.clear();
			extractInto = null;
			/* empty value means restore to the original location */
			if (value.equals("")) restoreMode = restoreType.origin;
			/* if contains an arrow assume directories have to be merged*/
			else if (value.contains("->")) {
				restoreMode = restoreType.mergeDirectory;
				List<String> pairs = new ArrayList<String>();
				/* multiple pairs separated by ; */
				if (value.contains(";")) pairs = Arrays.asList(value.split(";"));
				/* found a single entry */
				else pairs.add(value);
				/* for each source->destination pair */
				for (String pair : pairs) {
					final String[] match = pair.split("->");
					String source = match[0];
					String destination = match[1];
					/* we have a match */
					if (!source.equals("") && !destination.equals("")) {
						UserIO.print(UserIO.out.VERBOSE,"[CONFIG] merge "+ source+" with "+destination);    
						final FileUnit destinationFileUnit = new FileUnit(FileIO.getAbsolutePath(FileIO.neutralToNative(destination)),"",true);
						mergeDirectory.put(FileIO.nativeToNeutral(source),destinationFileUnit);
					}
				}
			} else {
				/* if there is a directory only */
				restoreMode = restoreType.extractInto;
				final FileUnit u = new FileUnit(FileIO.getAbsolutePath(value),"",true);
				extractInto = u;
				UserIO.print(UserIO.out.VERBOSE,"[CONFIG] extractInto directory = "+u.nativeName);
			}
			UserIO.print(UserIO.out.VERBOSE,"[CONFIG] restore mode = "+ restoreMode);
		}
		else UserIO.reportWarning(UserIO.out.INFO,"Skipping unknown configuration setting "+what);
	}

	/*
	 * returns a readable format of a given configuration setting
	 */
	public String setting2string(String what) {
		if (what.equals("directoriesToBackup")) {
			String directoriesToBackupString = "";
			for (int i = 0 ; i < sources.size(); i++) {
				directoriesToBackupString = directoriesToBackupString+sources.get(i).nativeName;
				if (i != sources.size()-1) directoriesToBackupString = directoriesToBackupString+";";
			}
			return sources.size() == 0 ? "" : directoriesToBackupString;
		}
		else if (what.equals("backupDirectory")) return archive != null ? archive.nativeName : "";
		else if (what.equals("password")) return password != null ? password : "";
		else if (what.equals("runAfterBackup")) return runAfterBackup != null ? runAfterBackup : "";
		else if (what.equals("purgeAfterDays")) return purgeAfterDays != -1 ? purgeAfterDays+"" : "";
		else if (what.equals("restoreInto")) {
			String restoreIntoString = "";
			if (restoreMode == restoreType.extractInto) restoreIntoString = extractInto.nativeName;
			else if (restoreMode == restoreType.mergeDirectory) {
				int i = 0;
				for (Entry<String, FileUnit> entry : mergeDirectory.entrySet()) {
					restoreIntoString = restoreIntoString+FileIO.neutralToNative(entry.getKey())+"->"+entry.getValue().nativeName;
					if (i != mergeDirectory.size()-1) restoreIntoString = restoreIntoString+";"; 
					i++;
				}
			}
			return restoreIntoString;
		}
		else if (what.equals("interactive")) return interactive ? "yes": "no" ;
		else if (what.equals("filter")) return filter != null ? filter.pattern() : "";
		else {
			UserIO.reportError("Invalid configuration setting requested for writing: "+what);
			return "";
		}
	}

	/*
	 * return the current hostname
	 */
	private String getHostname() {
		String hostname = null;
		try {
			hostname = InetAddress.getLocalHost().getHostName();
		} catch (Exception e) {
			UserIO.reportError("Unable to determine the hostname: " + UserIO.getExceptionMessage(e));
		} 
		if (hostname != null && hostname.contains(".")) hostname = hostname.split("\\.")[0];
		if (hostname == null || hostname.equals("")) hostname = "N.A.";
		return hostname;
	}

	/*
	 * Generate a template configuration file
	 */
	public void generate(File configFile, boolean overwrite) {
		UserIO.print(UserIO.out.INFO,"Requested action: SAVE JOB CONFIG...",false);
		/* ensure the config file does not exist yet */
		if (! overwrite && configFile.exists()) UserIO.reportError("A configuration file with the same name already exists at given location");
		try { 
			BufferedWriter bos = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(configFile)));
			bos.write("## ZipSnapNG Configuration file"+Main.lineSeparator);
			bos.write(""+Main.lineSeparator);
			bos.write("# directoriesToBackup: location on the filesystem to be included in the backup. Multiple sources can be added, separated by \";\" (e.g. c:\\dir1;c:\\dir2). Mandatory if command is \"backup\""+Main.lineSeparator);
			bos.write("directoriesToBackup = "+setting2string("directoriesToBackup")+Main.lineSeparator);
			bos.write("# backupDirectory: location on the filesystem where the backup will be stored. Mandatory for all the commands"+Main.lineSeparator);
			bos.write("backupDirectory = "+setting2string("backupDirectory")+Main.lineSeparator);
			bos.write(""+Main.lineSeparator);
			bos.write("# password: set an optional password for the backup/restore proccess (default: empty)"+Main.lineSeparator);
			bos.write("password = "+setting2string("password")+Main.lineSeparator);
			bos.write("# runAfterBackup: optional command to run after the backup. The backup directory will passed as first parameter, the backup file as second parameter (default: empty)"+Main.lineSeparator);
			bos.write("runAfterBackup = "+setting2string("runAfterBackup")+Main.lineSeparator);
			bos.write("# purgeAfterDays: optional number of days after which snapshots older than the threshold will be deleted. Pruning happens just after the backup process (default: empty)"+Main.lineSeparator);
			bos.write("purgeAfterDays =  "+setting2string("purgeAfterDays")+Main.lineSeparator);
			bos.write(""+Main.lineSeparator);
			bos.write("# restoreInto: instruct how to restore the snapshot. Available options are:"+Main.lineSeparator);
			bos.write("# - if empty, the files will be restored to the origin location"+Main.lineSeparator);
			bos.write("# - if a directory is provided, the files will be extracted there with the origin location converted into a folder (e.g. c:\\dir\\subdir\\file.txt -> c:\\restoreInto\\c-dir-subdir\\file.txt)"+Main.lineSeparator);
			bos.write("# - if source/destination mapping is provided separated by \"->\", the files will be restored by merging the original and destination directories configured in the mapping. Multiple pairs can be configured  separated by \";\" (e.g. c:\\original1,c:\\destination1;c:\\original2,c:\\destination2)"+Main.lineSeparator);
			bos.write("restoreInto = "+setting2string("restoreInto")+Main.lineSeparator);
			bos.write(""+Main.lineSeparator);
			bos.write("# interactive: require the interaction of the user. Set to \"no\" when running in a batch job (default: yes)"+Main.lineSeparator);
			bos.write("interactive = "+setting2string("interactive")+Main.lineSeparator);
			bos.write("# filter: optional regex filter to be applied for all the commands (default: empty)"+Main.lineSeparator);
			bos.write("filter = "+setting2string("filter")+Main.lineSeparator);
			bos.close();
		} catch (Exception e) {
			UserIO.reportError("Unable to generate job template file: " + UserIO.getExceptionMessage(e));
		} 
		UserIO.print(UserIO.out.INFO,"done");
		Job.exit(0);
	}
}