/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job;

import java.io.File;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.lingala.zip4j.core.ZipFile;
import net.lingala.zip4j.model.FileHeader;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.JobWrapper;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.unit.SliceUnit;
import com.backup.zipsnap.ng.job.util.FileIO;
import com.backup.zipsnap.ng.job.util.ProgressBar;
import com.backup.zipsnap.ng.job.util.UserIO;


public class Restore
{
	public static void doRestore() {
		UserIO.print(UserIO.out.INFO,"Requested action: RESTORE");
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[INITIALIZING]");    
		UserIO.print(UserIO.out.INFO,"Restore mode: "+Job.config.get().restoreMode.toString());
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"INITIALIZING");

		/* check if a catalog is found */
		if (Job.catalog.get().catalogFile == null) 
			UserIO.reportError("Unable to find a catalog in the specified archive directory " + Job.config.get().archive.nativeName);			

		/* check that catalog contains at least one snapshot */
		if (Job.catalog.get().snapshotsIndex.isEmpty()) 
			UserIO.reportError("No snapshots found in the catalog " + Job.catalog.get().catalogFile.getName());

		/*****************************************
		 * (1) IDENTIFY ITEMS TO RESTORE        *
		 ****************************************/
		/* display specified snapshot number and timestamp */
		UserIO.print(UserIO.out.INFO,"Snapshot number to restore: " +Job.config.get().snapshot);
		if (! Job.catalog.get().hasSnapshot(Job.config.get().snapshot)) 
			UserIO.reportError("No snapshot number "+Job.config.get().snapshot+" has not been found in the catalog");
		if (Job.catalog.get().snapshotsIndex.get(Job.config.get().snapshot-1).isDeleted) 
			UserIO.reportError("The snapshot number "+Job.config.get().snapshot+" has been deleted and cannot be restored");

		/* if encrypted, ensure a password is provided */
		if (Job.catalog.get().snapshotsIndex.get(Job.config.get().snapshot-1).isEncrypted && Job.config.get().password == null)
			UserIO.reportError("The snapshot #"+Job.config.get().snapshot+" is encrypted but no password has been provided");

		/* display the computer name of the snapshot */
		UserIO.print(UserIO.out.INFO,"Computer name: " +Job.catalog.get().snapshotsIndex.get(Job.config.get().snapshot-1).hostname);
		if (! Job.catalog.get().snapshotsIndex.get(Job.config.get().snapshot-1).hostname.equals(Job.config.get().hostname))
			UserIO.reportWarning(UserIO.out.INFO,"This computer name ("+Job.config.get().hostname+") is different than the computer from which the snapshot was taken (" +Job.catalog.get().snapshotsIndex.get(Job.config.get().snapshot-1).hostname+")");

		/* List directories in which the snapshot will be restored*/
		for (FileUnit directory : Job.config.get().directories) {
			UserIO.print(UserIO.out.INFO,"Destination directory: "+ directory.nativeName);
		}
		/* display filter, if defined */
		if (Job.config.get().filter != null) UserIO.print(UserIO.out.INFO,"Configured Regex filter: " + Job.config.get().filter.pattern());
		final List<FileUnit> snapshotFiles = new ArrayList<FileUnit>();
		for (FileUnit u : Job.catalog.get().files)
		{
			/* matching the filter and so to be restored */
			if (u.snapshotsIndex.contains(Job.config.get().snapshot-1) && u.matchesFilter()) {
				snapshotFiles.add(u);
			}
		}
		Job.memoryUsage("snapshotFiles",snapshotFiles);
		UserIO.print(UserIO.out.INFO,"Total items: " + snapshotFiles.size());

		/***************************************
		 * (2) SCAN DESTINATION DIRECTORIES     *
		 ***************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[ASSESSING DESTINATION DIRECTORY]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"ASSESSING DESTINATION DIRECTORY");
		Job.timer.get().start("ASSESSING DESTINATION DIRECTORY");
		/* get filtered contents of the current directory */
		final List<FileUnit> restoreDirContents = new ArrayList<FileUnit>();
		long dirSize = FileIO.listContents(Job.config.get().directories,restoreDirContents);
		Job.memoryUsage("restoreDirContents",restoreDirContents);
		Job.gui(JobWrapper.msg.TASK_LABEL,"Files in the destination directories: "+restoreDirContents.size()+" ("+FileIO.getSize(dirSize)+")");
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("ASSESSING DESTINATION DIRECTORY"));

		/***************************************
		 * (3) CALCULATE DIFFERENTIAL RESTORE *
		 **************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[CALCULATING DIFFERENTIAL RESTORE]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"CALCULATING DIFFERENTIAL RESTORE");
		Job.timer.get().start("CALCULATING DIFFERENTIAL RESTORE");
		/* Map for each snapshot/volume the files to be extracted */
		/* extractFiles[snapshot][volume] = FileUnit */
		final ArrayList<ArrayList<ArrayList<FileUnit>>> extractFiles = new ArrayList<ArrayList<ArrayList<FileUnit>>>(Job.config.get().snapshot);
		/* create the data structure first */
		for (int i = 0; i <= Job.config.get().snapshot-1; i++) {
			extractFiles.add(i, new ArrayList<ArrayList<FileUnit>>());
			for (int j = 0; j <= Job.catalog.get().snapshotsIndex.get(i).volumesIndex; j++) {
				extractFiles.get(i).add(j,new ArrayList<FileUnit>(Job.catalog.get().snapshotsIndex.get(i).volumesIndex+1));
			}
		}

		UserIO.print(UserIO.out.INFO,"Matching the catalog against the destination directories...",false);
		long partialSourceSize = 0;
		ProgressBar catalogMatchProgressBar = new ProgressBar();
		final List<FileUnit> createFiles = new ArrayList<FileUnit>();
		final List<FileUnit> updateFiles = new ArrayList<FileUnit>();

		/* for each file in the snapshot to be extracted */
		for (FileUnit u : snapshotFiles) {
			UserIO.print(UserIO.out.VERBOSE,"Evaluating "+u.nativeName+"...",false);
			FileUnit actualFile = null; /* reference the actual file on the filesystem */

			/* search for a file/directory in the destination directories */
			final int matchIndex = Collections.binarySearch(restoreDirContents, u);
			if (matchIndex < 0) {
				/* file not found on the filesystem */
				UserIO.print(UserIO.out.VERBOSE,"will be created",false);
				createFiles.add(u);     
			}  else {
				/* file already in the filesystem*/
				actualFile = restoreDirContents.get(matchIndex);
				/* calculates the checksum that will be used for file comparison */
				if (Main.settings.deepFileAnalysis) actualFile.updateChecksum();
				if ((u.size != actualFile.size) || (Math.abs(u.time - actualFile.time) > Main.settings.timeTolerance) || (!actualFile.checksum.equals("") && !u.checksum.equals(actualFile.checksum))) {
					UserIO.print(UserIO.out.ALL," not matching ("+
							u.size+"!="+ actualFile.size+
							" || "+u.time+" - "+actualFile.time+" > "+Main.settings.timeTolerance+
							" || "+u.checksum+" != "+actualFile.checksum+") ",false);
					UserIO.print(UserIO.out.VERBOSE,"will be updated",false);
					/* calculates the checksum if not done yet */
					if (! Main.settings.deepFileAnalysis) actualFile.updateChecksum();
					updateFiles.add(u);
				} else {
					UserIO.print(UserIO.out.VERBOSE,"will NOT be restored");
				}
			}

			/* if the item has to be extracted */
			if (createFiles.contains(u) || updateFiles.contains(u)) {
				if (u.isDirectory) {
					extractFiles.get(u.snapshotsIndex.get(0)).get(u.volume).add(u);
				} else {
					/* identify the slices to be restored */
					for (int i = 0; i < u.sliceList.size(); i++) {
						SliceUnit sliceToRestore = u.sliceList.get(i);
						sliceToRestore.snapshot = Job.catalog.get().slices.get(u.sliceList.get(i).checksum).snapshot;
						sliceToRestore.volume = Job.catalog.get().slices.get(u.sliceList.get(i).checksum).volume;
						/* if there is a corresponding slice on disk, check against the checksum */
						if (actualFile != null && (i+1) <= actualFile.sliceList.size()) {
							SliceUnit sliceOnDisk = actualFile.sliceList.get(i);
							if (! sliceOnDisk.checksum.equals(sliceToRestore.checksum)) sliceToRestore.isNew = true;
							else sliceToRestore.isNew = false;
						} 
						/* otherwise assume it is new (new file or file to restore bigger than the existing */
						else sliceToRestore.isNew = true;
						/* if the slice on the local drive is different then the original, restore it */
						if (sliceToRestore.isNew) {
							u.newSlices++;
							u.partialSize += Job.catalog.get().slices.get(sliceToRestore.checksum).size;
							/* add the file to the restore queue of the snapshot/volume */
							if (! extractFiles.get(sliceToRestore.snapshot-1).get(sliceToRestore.volume).contains(u)) {
								extractFiles.get(sliceToRestore.snapshot-1).get(sliceToRestore.volume).add(u);
							}
						}
					}
					UserIO.print(UserIO.out.VERBOSE," ["+u.newSlices+"/"+u.numSlices+"] ("+u.percentageNew()+"%)");
				}
			}

			partialSourceSize += u.size;
			catalogMatchProgressBar.display(partialSourceSize,dirSize);
		}
		Job.memoryUsage("extractFiles",extractFiles);
		UserIO.print(UserIO.out.INFO,"done");
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("CALCULATING DIFFERENTIAL RESTORE"));

		/********************************************
		 * (4) DETERMINE FILES TO DELETE           *
		 *******************************************/  
		List<FileUnit> deleteFiles = new ArrayList<FileUnit>();
		if (Main.settings.deleteAfterRestore) {
			UserIO.print(UserIO.out.INFO,"Determining files to be deleted...",false);
			/* (initialized to ALL files/directories in the current directory)      */
			deleteFiles = new ArrayList<FileUnit>(restoreDirContents);
			/* unmark file/directory to be deleted from the current directory */
			deleteFiles.removeAll(snapshotFiles);
			/* unmark file just created */
			deleteFiles.removeAll(createFiles);
			/* reverse sort the files/directories to be deleted */
			Collections.reverse(deleteFiles);
			Job.memoryUsage("deleteFiles",deleteFiles);
			UserIO.print(UserIO.out.INFO,"done");
		}

		/********************************************
		 * (5) PRINT THE RESTORE SUMMARY           , *
		 *******************************************/

		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[RESTORE SUMMARY]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"RESTORE SUMMARY");
		long createFilesSize = 0;
		long updateFilesSize = 0;
		long deleteFilesSize = 0;
		long restoreFilesSize = 0;
		int counter = 0;
		UserIO.out out = Main.settings.listFilesOnSummary ? UserIO.out.INFO : UserIO.out.DEBUG;

		if (createFiles.size() > 0) {
			counter = 0;
			UserIO.print(UserIO.out.INFO,"The following items will be CREATED: ",false);
			for (FileUnit u : createFiles) {
				if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
					UserIO.print(out,"- ...other "+(createFiles.size()-counter)+" items");	
					break;
				}
				String sizeString = u.size > 0 ? " ("+FileIO.getSize(u.size)+")" : "";
				UserIO.print(out,"- "+u.nativeName+sizeString);
				Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following items will be CREATED:",u.nativeName,sizeString});
				createFilesSize += u.partialSize;
				counter++;
			}
			UserIO.print(UserIO.out.INFO,createFiles.size()+ " items ("+FileIO.getSize(createFilesSize)+")");
		}

		if (updateFiles.size() > 0) {
			counter = 0;
			UserIO.print(UserIO.out.INFO,"The following items will be UPDATED: ",false);
			for (FileUnit u : updateFiles) {
				if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
					UserIO.print(out,"- ...other "+(updateFiles.size()-counter)+" items");	
					break;
				}
				String sizeString = u.size > 0 ? ": "+u.percentageNew()+"% ("+FileIO.getSize(u.partialSize)+"/"+FileIO.getSize(u.size)+")" : "";
				UserIO.print(out,"- "+u.nativeName+sizeString);
				Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following items will be UPDATE:",u.nativeName,sizeString});
				updateFilesSize += u.partialSize;
				counter++;
			}
			UserIO.print(UserIO.out.INFO,updateFiles.size()+ " items ("+FileIO.getSize(updateFilesSize)+")");
		}

		if (deleteFiles.size() > 0) {
			counter = 0;
			UserIO.print(UserIO.out.INFO,"The following items will be DELETED:",false);
			for (FileUnit u : deleteFiles) {
				if (Main.settings.summaryMaxFiles > 0 && counter >= Main.settings.summaryMaxFiles) {
					UserIO.print(out,"- ...other "+(deleteFiles.size()-counter)+" items");	
					break;
				}
				String sizeString = " ("+FileIO.getSize(u.size)+")";
				UserIO.print(out,"- "+u.nativeName+sizeString);
				Job.gui(JobWrapper.msg.TASK_FILE, new String[] {"The following items will be DELETED:",u.nativeName,sizeString});
				deleteFilesSize += u.size;
				counter++;
			}
			UserIO.print(UserIO.out.INFO,deleteFiles.size()+ " items ("+FileIO.getSize(deleteFilesSize)+")");

		}

		if (createFiles.size() == 0 && updateFiles.size() == 0 && deleteFiles.size() == 0) {
			/* nothing to be done */
			UserIO.print(UserIO.out.INFO,"No items need to be restored");
			Job.gui(JobWrapper.msg.NO_NEW_ITEMS, "");
			/* skip four steps */
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"RESTORING SNAPSHOT");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"VERIFYING RESTORED ITEMS");
			Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");
			return;   
		}
		
		Job.gui(JobWrapper.msg.SUMMARY_READY, "");
		restoreFilesSize = createFilesSize + updateFilesSize;
		UserIO.print(UserIO.out.INFO,"Total items to restore: "+(updateFiles.size()+createFiles.size())+" ("+FileIO.getSize(restoreFilesSize)+")");
		UserIO.print(UserIO.out.INFO,"Total items to delete: "+deleteFiles.size()+" ("+FileIO.getSize(deleteFilesSize)+")");
		Job.gui(JobWrapper.msg.TASK_LABEL,"Files to restore: "+(updateFiles.size()+createFiles.size())+" ("+FileIO.getSize(restoreFilesSize)+")  - Files to delete: "+deleteFiles.size()+" ("+FileIO.getSize(deleteFilesSize)+")");

		/* wait for user confirmation */
		if (Job.config.get().interactive) {
			Job.timer.get().pause("JOB");
			char choice = UserIO.userCharPrompt(UserIO.choice.YES_NO,"After reviewing the files which will be included in the snapshot, do you want to proceed with the restore?");
			if (choice == 'N') return;
			Job.timer.get().resume("JOB");
		}

		/***************************************************************
		 * (6) EXTRACT FILES/DIRECTORIES FROM THEIR RESPECTIVE VOLUMES *
		 ***************************************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[RESTORING SNAPSHOT]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"RESTORING SNAPSHOT");
		Job.timer.get().start("RESTORING SNAPSHOT");
		UserIO.print(UserIO.out.INFO,"Restore in progress...",false);
		/* successfully extracted files/directories */
		final List<FileUnit> extractedFiles = new ArrayList<FileUnit>();

		/* if destination directories do not exist, create them */
		for (FileUnit directory : Job.config.get().directories) {
			if (!directory.file.exists()) {
				UserIO.print(UserIO.out.DEBUG,"Creating destination directory "+directory.nativeName);
				directory.file.mkdirs();
			}
			if (!directory.file.isDirectory()) UserIO.reportError("Unable to create destination directory "+directory.nativeName);
		}

		/* extract files/directories from their respective volumes */
		long extractedFilesSize = 0;
		ProgressBar restoreProgressBar = new ProgressBar();
		ExtractNextSnapshot:
			/* for each snapshot */
			for (int snapshotIndex = 0; snapshotIndex <= Job.config.get().snapshot-1; snapshotIndex++) {
				final ArrayList<ArrayList<FileUnit>> extractFilesSnapshot = extractFiles.get(snapshotIndex);
				/* for each volume file */
				ExtractNextVolume:
					for (int volume = 0; volume < extractFilesSnapshot.size(); volume++) {
						/* files to be extracted for the given volume file*/
						final ArrayList<FileUnit> volumeFiles = extractFilesSnapshot.get(volume);
						if (volumeFiles.isEmpty()) continue ExtractNextVolume;
						UserIO.print(UserIO.out.DEBUG,"Extracting from snapshot "+(snapshotIndex+1)+":"+volume+"...");
						/* build the volume filename */
						String volumeName = FileIO.getCustomFilename("archive",(snapshotIndex + 1),Job.catalog.get().snapshotsIndex.get(snapshotIndex).time);
						if (volume > 0) volumeName = volumeName+ "."+volume;

						/* open the zip file */
						File zipFile = new File(Job.config.get().archive.file, volumeName + ".zip");
						if (zipFile == null || !zipFile.exists()) {
							UserIO.reportError("Unable to locate volume " + zipFile.getPath());
						}

						ZipFile zf = null;
						try {
							zf = new ZipFile(zipFile);
						} catch (Exception e) {
							UserIO.reportError("Unable to open ZIP volume " + zipFile.getPath()+ ": " + UserIO.getExceptionMessage(e));
						}
						/* if zip file is password protected then set the password */
						try {
							if (zf.isEncrypted()) {
								if (Job.config.get().password == null) UserIO.reportError("The ZIP volume " + zipFile.getPath()+ " is encypted but now password is provided");
								zf.setPassword(Job.config.get().password);
							}
						} catch (Exception e) {
							UserIO.reportWarning(UserIO.out.INFO,"Unable to determine if the ZIP volume " + zipFile.getPath()+ " is password protected: " + UserIO.getExceptionMessage(e));
						}

						/* for each file supposed to be extracted from the volume */
						ExtractNextFile:
							for (FileUnit u : volumeFiles) {
								boolean extractFile = false;
								/* Check if the file exists on the local filesystem */
								String restoreFrom = (u.restoreFrom != null) ? " (origin "+u.restoreFrom.nativeName+")"  : "";
								/* if the file already exists determine if it has to be overwritten */
								if (u.file.exists()) {
									/* a file/directory of the same name already exists */
									if (u.isDirectory == u.file.isDirectory()) {
										char choice = ' ';
										/* overwrite existing file/directory? */
										if (Main.settings.overwriteExisting.equals("prompt")) {
											if (Job.config.get().directories.contains(u)) choice = 'Y';
											else {
												choice = UserIO.userCharPrompt(UserIO.choice.YES_NO_ALWAYS_NEVER,"Do you want to overwrite \"" + u.nativeName + "\"? ");
											}
										} 
										if (choice == 'A') Main.settings.overwriteExisting = "yes";
										if (choice == 'R') Main.settings.overwriteExisting = "no";
										if (choice == 'Y' || Main.settings.overwriteExisting.equals("yes")) {
											UserIO.print(UserIO.out.DEBUG,"Overwriting " + u.nativeName+restoreFrom,false);
											extractFile = true;
										} 
										if (choice == 'N' || Main.settings.overwriteExisting.equals("no")) {
											UserIO.print(UserIO.out.DEBUG,"Skipping existing " + u.nativeName+restoreFrom);
											extractFile = true;
										}        
									}
									else {
										UserIO.reportWarning(UserIO.out.INFO,"Unable to extract " + u.nativeName + "\" from volume \"" + zipFile.getPath() + ": An item with the same name already exists");
										continue ExtractNextFile;
									}
								} else {
									/* file/directory does not exist yet */
									UserIO.print(UserIO.out.DEBUG,"Creating " + u.nativeName+restoreFrom,false);
									extractFile = true;
								}
								if (!extractFile) continue ExtractNextFile;

								/* extract a directory */
								if (u.isDirectory) {
									u.file.mkdirs();
									if (!u.file.exists() || !u.file.isDirectory()) {
										/* something went wrong */
										UserIO.reportWarning(UserIO.out.INFO,"Unable to extract directory " + u.nativeName + " from ZIP volume " + zipFile.getPath());
										continue ExtractNextFile;
									}
									extractedFiles.add(u);
									UserIO.print(UserIO.out.DEBUG,"");
									continue ExtractNextFile;
								}
								/* extract a file */
								else {
									/* create parent directory if does not exist */
									final File parentDir = u.file.getParentFile();
									if ((parentDir != null) && !parentDir.exists()) parentDir.mkdirs();
									if ((parentDir == null) || !parentDir.isDirectory()) {
										UserIO.reportWarning(UserIO.out.INFO,"Unable to extract file "+u.nativeName + " from ZIP volume " + zipFile.getPath() +
												": The parent directory of the file does not exist and cannot be created");
										continue ExtractNextFile;
									}

									/* for each slice of the file */
									RandomAccessFile ptr = null;
									UserIO.print(UserIO.out.DEBUG," [",false);
									ExtractNextSlice:
										for (int i = 0; i < u.sliceList.size(); i++) {
											SliceUnit slice = u.sliceList.get(i);
											/* if the slice is not in this volume, skip it */
											slice.snapshot = Job.catalog.get().slices.get(u.sliceList.get(i).checksum).snapshot;
											slice.volume = Job.catalog.get().slices.get(u.sliceList.get(i).checksum).volume;
											if (slice.snapshot != (snapshotIndex+1) || slice.volume != volume) {
												UserIO.print(UserIO.out.DEBUG," ",false);
												continue ExtractNextSlice;
											}
											UserIO.print(UserIO.out.ALL,"("+slice.checksum+")",false);
											/* if the slice is not to be replaced, skip it*/
											if (!slice.isNew) {
												UserIO.print(UserIO.out.DEBUG,".",false);
												continue ExtractNextSlice;
											}
											UserIO.print(UserIO.out.DEBUG,"N",false);       
											/* open the slice from the zip file and get the stream */
											FileHeader fileHeader = null;
											try {
												fileHeader = zf.getFileHeader(slice.checksum);
											} catch (Exception e) {
												UserIO.reportError("Unable to retrieve file header for "+slice.checksum+" in ZIP volume " + zipFile.getName() + ": " +UserIO.getExceptionMessage(e));
											}
											if (fileHeader == null) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to extract slice "+slice.checksum+" belonging to the file " +u.nativeName + " from ZIP volume " + zipFile.getPath() +
														": Cannot find the corresponding entry in the ZIP volume");
												continue ExtractNextFile;
											}
											InputStream zis = null;
											try {
												zis = zf.getInputStream(fileHeader);
											} catch (Exception e) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to read slice "+slice.checksum+" belonging to the file " +u.nativeName + " from ZIP volume " + zipFile.getPath() +
														": Cannot find the corresponding entry in the ZIP volume");
												continue ExtractNextFile;
											} 
											/* open the destination file for writing */
											try {
												ptr = new RandomAccessFile(u.file, "rw");
											} catch (Exception e) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to open file "+u.file.getPath()+" for writing: " + UserIO.getExceptionMessage(e));
												continue ExtractNextFile;
											} 

											/* copy the content of the slice into the destination file */
											final byte byteBuffer[] = new byte[Main.settings.bufferSize];
											try {
												/* seek the beginning of the slice */
												ptr.seek(i*u.sliceSize);
												while (true) {
													/* read the source file */
													final int byteCount = zis.read(byteBuffer, 0, Main.settings.bufferSize);
													if (byteCount == -1) break; /* reached EOF */
													/* write to the destination file */
													ptr.write(byteBuffer,0,byteCount);
												}
												/* if this is the last slide, set the size */ 
												if (i == u.sliceList.size()-1) ptr.setLength(u.size);
											} catch (Exception e) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to extract slice "+(i+1)+" of " + u.nativeName +" from ZIP volume " + zipFile.getPath() + ": " +
														UserIO.getExceptionMessage(e));
												continue ExtractNextFile;
											} 
											/* close the input and output streams */
											try {
												zis.close();
											} catch (Exception e) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to close file ZIP file "+zipFile.getPath()+" after restoring: "+ UserIO.getExceptionMessage(e));
												continue ExtractNextFile;
											}
											try {
												ptr.close();
											} catch (Exception e) {
												UserIO.reportWarning(UserIO.out.INFO,"Unable to close file "+u.nativeName+" after restoring: "+ UserIO.getExceptionMessage(e));
												continue ExtractNextFile;
											}
											extractedFilesSize += slice.size;
											/* Show progress */
											restoreProgressBar.display(extractedFilesSize,restoreFilesSize);
										} /* end for each slice */

									/* set the file timestamp */
									final FileIO.FileIOResult result = FileIO.setFileTime(u.file, u.time);
									if (!result.success) UserIO.reportWarning(UserIO.out.INFO,"Unable to set last-modified time of extracted item " + u.nativeName);
								} /* extract a file */

								/* file/directory successfully extracted from the ZIP volume */
								extractedFiles.add(u);
								UserIO.print(UserIO.out.DEBUG,"]",false);
								UserIO.print(UserIO.out.DEBUG,"");
								continue ExtractNextFile;    
							} /* end for each file */
						continue ExtractNextVolume;
					} /* end for each volume */
				continue ExtractNextSnapshot;
			} /* end for each snapshot */
		UserIO.print(UserIO.out.INFO,"done");
		Job.memoryUsage("extractedFiles",extractedFiles);
		UserIO.print(UserIO.out.INFO,"Extracted " + extractedFiles.size() +" items ("+FileIO.getSize(extractedFilesSize)+")");

		/*********************
		 * (7) DELETE FILES *
		 ********************/
		final List<FileUnit> deletedFiles = new ArrayList<FileUnit>();
		if (deleteFiles.size() > 0) {
			UserIO.print(UserIO.out.INFO,"Deleting files...",false);
			long deletedFilesSize = 0;
			ProgressBar deleteProgressBar = new ProgressBar();

			DeleteNextFile:
				for (FileUnit u : deleteFiles) {
					if (!u.file.exists() || (u.isDirectory != u.file.isDirectory())) {
						UserIO.reportWarning(UserIO.out.INFO,u.nativeName + " does not exist anymore");
						continue DeleteNextFile;
					}
					/* get canonical pathname of file/directory to be deleted */
					String pathname = null;
					try { pathname = u.file.getCanonicalPath();
					} catch (Exception e) {
						pathname = null;
					}
					if ((pathname == null) || !pathname.equals(u.file.getPath())) {
						UserIO.reportWarning(UserIO.out.INFO,u.nativeName + " does not exist anymore");
						continue DeleteNextFile;
					}

					boolean deleteFile = false;
					if (Main.settings.deleteExisting.equals("yes")) {
						UserIO.print(UserIO.out.DEBUG,"Deleting " + u.nativeName);
						deleteFile = true;
					} else if (Main.settings.deleteExisting.equals("no")) {
						UserIO.print(UserIO.out.DEBUG,"Skipping " + u.nativeName);
					} else if (Main.settings.deleteExisting.equals("prompt")) {
						final char choice = UserIO.userCharPrompt(UserIO.choice.YES_NO_ALWAYS_NEVER,"Do you want to delete " + u.nativeName + "? ");
						if (choice == 'Y') {
							deleteFile = true;
						} else if (choice == 'A') {
							Main.settings.deleteExisting = "yes";
							deleteFile = true;
						} else if (choice == 'R') {
							Main.settings.deleteExisting = "no";
						}
					}
					if (! deleteFile) continue DeleteNextFile;
					/* delete the file */
					FileIO.FileIOResult result = null;
					if (u.isDirectory) {
						result = FileIO.deleteDir(u.file);
					} else {
						result = FileIO.deleteFile(u.file);
					}
					if (result.success) {
						deletedFiles.add(u);
						deletedFilesSize += u.size;
						/* Show progress */
						deleteProgressBar.display(deletedFilesSize,deleteFilesSize);
					} else {
						UserIO.reportWarning(UserIO.out.INFO,"Unable to delete "+u.nativeName + ": " + result.errorMessage);
					}
				}
			UserIO.print(UserIO.out.DEBUG,"Deleted "+deletedFiles.size()+" items ("+FileIO.getSize(deletedFilesSize)+")");
			UserIO.print(UserIO.out.INFO,"done");
			Job.memoryUsage("deletedFiles",deletedFiles);
		}
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("RESTORING SNAPSHOT"));

		/***************************************
		 * (8) RESTORE DIRECTORY TIMESTAMPS  *
		 **************************************/
		UserIO.print(UserIO.out.INFO,Main.lineSeparator+"[FINALIZING]");
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"FINALIZING");
		Job.timer.get().start("FINALIZING");
		UserIO.print(UserIO.out.INFO,"Restore directory timestamps...",false);
		/* updated current directory contents, initialized as the current dir contents */
		final List<FileUnit> updatedRestoreDirContents = new ArrayList<FileUnit>(restoreDirContents);

		/* update the restore dir contents */
		for (FileUnit u : extractedFiles) {
			final int i = Collections.binarySearch(restoreDirContents, u);
			if (i >= 0) {
				/* overwrite existing file/directory */
				updatedRestoreDirContents.set(i, u);
			} else {
				/* add extracted file/directory */
				updatedRestoreDirContents.add(u);
			}
		}
		Collections.sort(updatedRestoreDirContents);
		updatedRestoreDirContents.removeAll(deletedFiles);
		Job.memoryUsage("updatedRestoreDirContents",updatedRestoreDirContents);

		/* determine directories with timestamps to be restored */
		final List<FileUnit> restoreDirs = new ArrayList<FileUnit>();
		final List<FileUnit> restoredDirs = new ArrayList<FileUnit>();
		long restoreSize = 0;
		for (FileUnit u : updatedRestoreDirContents) {
			restoreSize += u.size;
			if (u.isDirectory && u.file.isDirectory() && (u.file.lastModified() != u.time)) {
				/* get canonical pathname of directory */
				String pathname = null;
				try {
					pathname = u.file.getCanonicalPath();
				} catch (Exception e) {
					pathname = null;
				}
				if ((pathname != null) && pathname.equals(u.file.getPath())) restoreDirs.add(u);
			}
		}
		Collections.reverse(restoreDirs);

		/* restore directory timestamps in the current directory */
		if (!restoreDirs.isEmpty()) {
			for (FileUnit u : restoreDirs) {
				UserIO.print(UserIO.out.VERBOSE,"Restoring timestamp of " + u.nativeName);
				final FileIO.FileIOResult result = FileIO.setDirTime(u.file, u.time);
				if (result.success) {
					restoredDirs.add(u);
				} else {
					UserIO.reportWarning(UserIO.out.INFO,"Unable to restore timestamp of directory " + u.nativeName);
				}
			}
		}
		UserIO.print(UserIO.out.INFO,"done");

		/***************************************
		 * (9) VERIFY RESTORE                 *
		 **************************************/    
		Job.gui(JobWrapper.msg.JOB_PROGRESS,"VERIFYING RESTORED ITEMS");
		if (Main.settings.verifyAfterRestore) {
			ProgressBar verifyProgressBar = new ProgressBar();
			UserIO.print(UserIO.out.INFO,"Verifying restored items...",false);
			/* for each file that has been restored */
			long restoredSize = 0;
			for (FileUnit u : extractedFiles) {
				FileUnit restoredFile = new FileUnit(u.nativeName,"",u.isDirectory);
				/* update its properties (timestamp and size) */
				restoredFile.update();
				/* compute the checksum */
				restoredFile.updateChecksum();
				/* check if everything has been restored correctly*/
				if (u.time != restoredFile.time) 
					UserIO.reportWarning(UserIO.out.INFO,"The item "+u.nativeName+" timestamp has not been restored correctly (expecting "+u.time+" but found "+restoredFile.time+")");
				if (! u.isDirectory && ! u.checksum.equals(restoredFile.checksum)) 
					UserIO.reportWarning(UserIO.out.INFO,"The item "+u.nativeName+" checksum does not match (expecting "+u.checksum+" but found "+restoredFile.checksum+"). The file may have not restored correctly");
				restoredSize += u.size;
				verifyProgressBar.display(restoredSize,restoreSize);
			}
			UserIO.print(UserIO.out.INFO,"done");
		}

		Job.gui(JobWrapper.msg.JOB_PROGRESS,"DONE");
		UserIO.print(UserIO.out.DEBUG,"Elapsed: "+Job.timer.get().stop("FINALIZING"));
	}

}