/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.security.MessageDigest;
import java.text.Normalizer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.regex.Pattern;
import java.util.zip.CRC32;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.job.Job;
import com.backup.zipsnap.ng.job.unit.FileUnit;
import com.backup.zipsnap.ng.job.unit.SliceUnit;

/**
 * Simple class for performing common file input/output operations.
 */
public class FileIO
{
  
  /********************************
    * FILE I/O FUNCTIONS          *
    *******************************/
  
  /**
   * Set last-modified time of a given file or directory.
   */
  private static FileIOResult setFileDirTime(final boolean isDir, final File file, final long time) {
    /* check if file/directory exists */
    if (!file.exists()) return new FileIOResult(false,((isDir) ? "Directory" : "File") + " does not exist.");
    /* check if file/directory is a file/directory as specified */
    if (isDir != file.isDirectory()) return new FileIOResult(false,
                                                             "Specified " + ((isDir) ? "directory" : "file") +
                                                             " is not a " + ((isDir) ? "directory." : "file."));
    boolean success = false;
    String exceptionMessage = null;
    try {
      success = file.setLastModified(time);
    } catch (Exception e) {
      exceptionMessage = UserIO.getExceptionMessage(e);
    }
    if (success) {
      return new FileIOResult(true, null);
    } else {
      if (exceptionMessage == null) {
        return new FileIOResult(false,
                                "Setting of last-modified time of " +
                                ((isDir) ? "directory" : "file") + " was unsuccessful.");
      } else {
        return new FileIOResult(false,
                                "Java exception encountered during " + "setting of last-modified time of " +
                                ((isDir) ? "directory: " : "file: ") + exceptionMessage);
      }
    }
  }
  
  /**
   * Set last-modified time of a given file.
   */
  public static FileIOResult setFileTime(final File file,final long time)
  {
    return setFileDirTime(false, file, time);
  }
  
  /**
   * Set last-modified time of a given directory.
   */
  public static FileIOResult setDirTime(final File dir,final long time)
  {
    return setFileDirTime(true, dir, time);
  }
  
  
  /**
   * Create directory.
   */
  public static FileIOResult createDir(final File dir) {
    boolean success = false;
    String exceptionMessage = null;
    try {
      success = dir.mkdirs();
    } catch (Exception e) {
      exceptionMessage = UserIO.getExceptionMessage(e);
    }
    if (success) {
      return new FileIOResult(true, null);
    }
    else {
      if (exceptionMessage == null) {
        return new FileIOResult(false,"Directory creation was unsuccessful.");
      }else {
        return new FileIOResult(false,"Java exception encountered during directory creation: " + exceptionMessage);
      }
    }
  }
  
  /**
   * Delete a file or directory.
   */
  private static FileIOResult deleteFileDir(final boolean isDir,final File file) {
    /* check if specified file/directory exists */
    if (!file.exists()) return new FileIOResult(false,((isDir) ? "Directory" : "File") + " does not exist.");
    /* check if file/directory is a file/directory as specified */
    if (isDir != file.isDirectory()) return new FileIOResult(false,
                                                             "Specified " + ((isDir) ? "directory" : "file") +
                                                             " is not a " + ((isDir) ? "directory." : "file."));
    boolean success = false;
    String exceptionMessage = null;
    try {
      success = file.delete();
    } catch (Exception e) {
      exceptionMessage = UserIO.getExceptionMessage(e);
    }
    if (success) {
      return new FileIOResult(true, null);
    } else {
      if (exceptionMessage == null) {
        return new FileIOResult(false,"Deleting of " + ((isDir) ? "directory" : "file") + " was unsuccessful.");
      } else       {
        return new FileIOResult(false,"Java exception encountered during deleting of " +
                                ((isDir) ? "directory: " : "file: ") + exceptionMessage);
      }
    }
  }
  
  /**
   * Delete a file.
   * 
   */
  public static FileIOResult deleteFile(final File file)
  {
    return deleteFileDir(false, file);
  }
  
  /**
   * Delete a directory.
   */
  public static FileIOResult deleteDir(final File dir)
  {
    return deleteFileDir(true, dir);
  }
  
  
  /**
   * Copy a file.
   */
  public static FileIOResult copyFile(final File sourceFile,final File targetFile)
  {
    /* buffered input stream for reading */
    BufferedInputStream bis = null;
    /* buffered output stream for writing */
    BufferedOutputStream bos = null;
    try {
      /* error messages, if any */
      final StringBuilder errorMessages = new StringBuilder();
      try {
        bis = new BufferedInputStream(new FileInputStream(sourceFile));
      } catch (Exception e) {
        errorMessages.append("Unable to open source file for reading: " +
                             UserIO.getExceptionMessage(e));
        return new FileIOResult(false, errorMessages.toString());
      }
      /* parent directory of the target file */
      final File targetParentDir = targetFile.getParentFile();
      /* create parent directory of target file, if necessary */
      if (!targetParentDir.exists()) createDir(targetParentDir);
      if (!targetParentDir.isDirectory()) {
        errorMessages.append("Unable to create parent directory of target file.");
        return new FileIOResult(false, errorMessages.toString());
      }
      try {
        bos = new BufferedOutputStream(new FileOutputStream(targetFile));
      } catch (Exception e) {
        errorMessages.append("Unable to open target file for writing: " +UserIO.getExceptionMessage(e));
        return new FileIOResult(false, errorMessages.toString());
      }
      /* byte buffer */
      final byte byteBuffer[] = new byte[Main.settings.bufferSize];
      try {
        /* copy bytes from the source file to the target file */
        while (true) {
          final int byteCount = bis.read(byteBuffer, 0, Main.settings.bufferSize);
          if (byteCount == -1) break; /* reached EOF */
          bos.write(byteBuffer, 0, byteCount);
        }
      } catch (Exception e) {
        errorMessages.append("Unable to copy data from source file to target file: " + UserIO.getExceptionMessage(e));
        return new FileIOResult(false, errorMessages.toString());
      }
      try {
        bis.close();
        bis = null;
      } catch (Exception e) {
        errorMessages.append("Unable to close source file after reading: ");
        errorMessages.append(UserIO.getExceptionMessage(e));
      }
      try {
        bos.close();
        bos = null;
      } catch (Exception e) {
        errorMessages.append("Unable to close target file after writing: ");
        errorMessages.append(UserIO.getExceptionMessage(e));
      }
      final FileIOResult setFileTimeResult = setFileTime(targetFile, sourceFile.lastModified());
      if (!setFileTimeResult.success)  {
        errorMessages.append("Unable to set last-modified time of target file:");
        errorMessages.append(setFileTimeResult.errorMessage);
      }
      /* result of file copy operation */
      return new FileIOResult(setFileTimeResult.success, errorMessages.toString());
    }finally {
      /* close buffered input stream for reading */
      if (bis != null) {
        try {
          bis.close();
        } catch (Exception e) {
          /* ignore */
        }
      }
      /* close buffered output stream for writing */
      if (bos != null) {
        try {
          bos.close();
        } catch (Exception e) {
          /* ignore */
        }
      }
    }
  }
  
  /**
   * Return the checksum of this file/directory.
   */
  public static ChecksumResult calculateChecksum(FileUnit u) {
    ArrayList<SliceUnit> sliceList = new ArrayList<SliceUnit>();
    final StringBuilder errorMessages = new StringBuilder();
    /* checksum of directory is defined as 0 */
    if (u.file.isDirectory()) {
      return new ChecksumResult(true, null, "",sliceList);
    }
    /* open the file */
    RandomAccessFile ptr = null;
    try {
      ptr = new RandomAccessFile(u.file, "r");
    } catch (Exception e) {
      errorMessages.append("cannot open " + UserIO.getExceptionMessage(e));
      return new ChecksumResult(false, errorMessages.toString(), "",sliceList);
    }
    /* setup checksums */
    String algorithm = "SHA-256";
    MessageDigest md = null;
    MessageDigest mdSlice = null;
    CRC32 crc32Slice = new CRC32();
    try {
      md = MessageDigest.getInstance(algorithm);
      mdSlice = MessageDigest.getInstance(algorithm);
    } catch (Exception e) {
      UserIO.reportError("Hashing algorithm "+algorithm+" not supported");
    }
    int sliceIndex = 0;
    /* read bytes from the file, and track the checksum computation */
    int sliceSize = 0;
    while (true) {
      /* allocate a buffer  */
      byte byteBuffer[] = new byte[Main.settings.bufferSize];
      /* ensure byteBuffer is fully filled in */
      int read = fillBuffer(u,ptr,byteBuffer);
      /* an error has occurred */
      if (read == -1) {
        errorMessages.append("cannot read "+u.nativeName);
        return new ChecksumResult(false, errorMessages.toString(), "",sliceList);
      }
      /* record the bytes read so far */
      sliceSize += read;
      /* update both the file and slice checsum */
      if (read > 0) {
        md.update(byteBuffer, 0, read);
        mdSlice.update(byteBuffer, 0, read);
        crc32Slice.update(byteBuffer, 0, read);
      }
      /* time to record the crc of the current slice*/
      if ((sliceSize == u.sliceSize) || read == 0) {
        SliceUnit slice = new SliceUnit(bytesToHex(mdSlice.digest()));
        slice.size = sliceSize;
        slice.crc = crc32Slice.getValue();
        sliceList.add(sliceIndex,slice);
        sliceIndex++;
        sliceSize = 0;
        crc32Slice.reset();
        mdSlice.reset();
      }
      /* no more bytes to read */
      if (read == 0) break;
    }
    /* close the file */
    try {
      ptr.close();
      ptr = null;
    } catch (Exception e)  {
      errorMessages.append("cannot close "+ UserIO.getExceptionMessage(e));
      return new ChecksumResult(false, errorMessages.toString(), "",sliceList);
    }
    return new ChecksumResult(true, errorMessages.toString(),bytesToHex(md.digest()),sliceList);
  }
  
  
  /**
   * Convert a byte array into a Hex string
   */  
  public static String bytesToHex(byte[] bytes) {
    StringBuffer result = new StringBuffer();
        for (byte byt : bytes) result.append(Integer.toString((byt & 0xff) + 0x100, 16).substring(1));
        return result.toString();
    }
    
  
  
  /**
   * ensure the result array is filled before returning it while reading from a file
   */
  public static int fillBuffer(FileUnit u, RandomAccessFile ptr, byte[] result) {
    /* buffer to store the current read cycle */
    int size = result.length;
    final byte byteBuffer[] = new byte[size];
    /* bytes read so far */
    int byteRead = 0;
    /* until the entire buffer size is read */
    while(byteRead < size) {
      try {
        /* attempt to read the entire buffer size */
        int byteCount = ptr.read(byteBuffer, 0, size);
        if (byteCount == -1) {
          return byteRead;
        }
        /* by default the entire buffer lenght will be copied to the result array */
        int length = byteCount;
        /* if we are readong too much, cut the buffer to the remaining bytes of the result array */
        if (byteRead+byteCount > size) length = size - byteRead;
        /* store into the result buffer what has been read */
        System.arraycopy(byteBuffer,0,result,byteRead,length);
        byteRead += byteCount;
      } catch (Exception e) {
        UserIO.reportWarning(UserIO.out.DEBUG,"An unexpected error has occurred while reading "+u.nativeName+": " +UserIO.getExceptionMessage(e));
        return -1;
      }
    }
    return byteRead;
  }
  
  /**
   * check the validity of a given directory
   */
  public static void checkDirectory(FileUnit directory) {
    //if (directory == null || directory.file == null) UserIO.reportError("The directory provided is invalid");
    if (!directory.file.exists())
      UserIO.reportError("The specified directory " + directory.nativeName + " does not exist.");
    if (!directory.file.isDirectory())
      UserIO.reportError("The specified directory " + directory.nativeName + " is not a directory; could it be a file?");
    if (!directory.file.canRead())
      UserIO.reportError("The specified directory " + directory.nativeName + " cannot be read.");
  }
  
  /**
   * check the validity of a given file and create its parent directory if it does not exist
   */
  public static void checkFile(File file) {
    final File parentDir = file.getParentFile();
    /* create parent directories if necessary */
    if ((parentDir != null) && !parentDir.exists()) parentDir.mkdirs();
    
    /* check that the parent directory of the ZIP volume exists */
    if (!parentDir.isDirectory()) UserIO.reportError("Unable to create the parent directory " + parentDir.getPath());
    if (file.exists()) UserIO.reportError( "Unable to create file " + file.getName() + ": A file of the same name already exists");
  }
  
  /**
   * Return the absolute path of a given file
   */
  public static String getAbsolutePath(File file) {
    String fullPath = "";
    try {
      fullPath = file.getCanonicalPath();
    } catch (Exception e) {
      UserIO.reportError("Unable to get absolute path of " + file.getPath() + ": " +UserIO.getExceptionMessage(e));
    }
    return fullPath;
  }
  
  /**
   * Return the absolute path of a given file
   */
  public static String getAbsolutePath(String filename) {
    File file = new File(filename);
    return getAbsolutePath(file);
  }
  
  /********************************
    * STRING MANIPULATION         *
    *******************************/
  
  /**
   * Convert a path string from native form to neutral form.
   * In neutral form, the separator is always FileIO.neutralSeparatorChar.
   */
  public static String nativeToNeutral(final String path)
  {
    return path.replace(Main.fileSeparator, Main.neutralSeparator);
  }
  
  
  /**
   * Convert a path string from neutral form to native form.
   * In neutral form, the separator is always FileIO.neutralSeparatorChar.
   */
  public static String neutralToNative(final String path)
  {
    return path.replace(Main.neutralSeparator, Main.fileSeparator);
  }
  
  
  /**
   * Removes a trailing separator, if any, in the specified path string.
   */
  public static String trimTrailingSeparator(final String path) {
    if (path.endsWith(File.separator)) return path.substring(0, path.length() - 1);
    return path;
  }
  
  /**
   * Removes a leading separator, if any, in the specified path string.
   */
  public static String trimLeadingSeparator(final String path) {
    if (path.startsWith(File.separator)) return path.substring(1, path.length());
    return path;
  }
  
  /**
   * convert a path into a string
   */
  public static String getNormalizedPath (final String path) {
    if (path == null) return "";
    String separator = "-";
    String res = "";
    res = path;
    res = res.replaceAll(Pattern.quote(":"),"");
    res = res.replaceAll(Pattern.quote(Main.fileSeparator),separator);
    res = res.replaceAll(Pattern.quote(Main.neutralSeparator),separator);
    return res;
  }
  
  
  /**
   * Return human readable strings for bytes.
   */
  public static String getSize(long bytes) {
    boolean si = false;
    if (bytes < 0) bytes = 0;
    int unit = si ? 1000 : 1024;
    if (bytes < unit) return bytes + " B";
    int exp = (int) (Math.log(bytes) / Math.log(unit));
    String pre = (si ? "kMGTPE" : "KMGTPE").charAt(exp-1) + (si ? "" : "");
    return String.format("%.1f %sB", bytes / Math.pow(unit, exp), pre);
  }
  
  /**
   * Return an archive file name
   */
  public static String getCustomFilename(String type) {
    String prefix = "zipsnap."+Job.config.get().jobName + "." +Job.config.get().snapshot+ "." + Job.timestamp.get();
    if (type.equals("")) return prefix;
    return "zipsnap."+Job.config.get().jobName + "." +Job.config.get().snapshot+ "." + Job.timestamp.get()+"."+type;
  }
  public static String getCustomFilename(String type,int snapshot,String timestamp) {
    return "zipsnap."+Job.config.get().jobName + "." +snapshot+ "." + timestamp+"."+type;
  }
  
  /**
   * Return the percentage of current over total
   */  
  public static int percentage(long current, long total) {
    long percentage = 0;
    if (total == 0) return 0;
    percentage = current * 100 / total;
    if (percentage < 0) percentage = 0;
    Long percentageLong = new Long(percentage);
    return percentageLong.intValue();
  }
  
  /**
   * Return the compressio ratio
   */  
  public static int compressionRatio(long uncompressSize, long compressSize) {
    long ratio = 0;
    if (uncompressSize <= 0) return 0;
    ratio = (uncompressSize - compressSize) * 100 / uncompressSize;
    if (ratio < 0) return 0;
    Long ratioLong = new Long(ratio);
    return ratioLong.intValue();
  }
  
  /**
   * convert a crc-32 into a string and vice versa
   */ 
  public static String crc2string(Long crc) {
    return Long.toHexString(crc).toUpperCase(Locale.ENGLISH);
  }
  public static Long string2crc(String crc) {
    return Long.parseLong(crc, 16);
  }
  
  /**
   * Convert a real path into a shadow path
   */  
  public static String real2shadow(String real) {
    if (Job.config.get().shadowDevice == null) return real;
    return real.replaceFirst("^\\w:",Job.config.get().shadowDevice);
  }
  
  /**
   * Convert a shadow path into a real path
   */  
  public static String shadow2real(String shadow) {
    if (Job.config.get().shadowDevice == null) return shadow;
    return shadow.replaceFirst("^"+Pattern.quote(Job.config.get().shadowDevice),Job.config.get().shadowVolume+":");
  }
  
  
  /**
   * Normalize to "Normalization Form Canonical Decomposition" (NFD)
   */
  public static String normalizeUnicode(String str) {
    Normalizer.Form form = Normalizer.Form.NFD;
    if (!Normalizer.isNormalized(str, form)) {
      return Normalizer.normalize(str, form);
    }
    return str;
  }
  
  /**
   * Normalize a string into a a given encoding
   */
  public static String normalizeEncoding(String str) {
    String s = null;
    String encoding = "ISO-8859-1";
    try {
      s =  new String (str.getBytes("UTF-8"),encoding);
    } catch (Exception e){
      UserIO.reportWarning(UserIO.out.INFO,"Encoding "+encoding+" not supported: " + UserIO.getExceptionMessage(e));
    }
    return s;    
  }

  /*
   * determine if a file is a symlink
   */
  public static boolean isSymlink(File file)  {
	  File canon;
	  boolean isSymlink = false;
	  try {
		  if (file.getParent() == null) {
			  canon = file;
		  } else {
			  File canonDir = file.getParentFile().getCanonicalFile();
			  canon = new File(canonDir, file.getName());
		  }
		  isSymlink = !canon.getCanonicalFile().equals(canon.getAbsoluteFile());
		  if (isSymlink) {
			  UserIO.print(UserIO.out.DEBUG,"Symlink detected for "+file.getPath()+" (" + canon.getAbsoluteFile()+" -> "+canon.getCanonicalFile()+")");
		  }
	  } catch (Exception e){
		  UserIO.reportWarning(UserIO.out.INFO,"Error while determining if "+file.getPath()+" is a symlink: " + UserIO.getExceptionMessage(e));
	  }
	  return isSymlink;
  }
  
  /********************************
    * GET DIR CONTENTS            *
    *******************************/
  
  public static long listContents(List<FileUnit> directories,List<FileUnit> dirContents) {
    int dirCount = 0;
    long dirSize = 0;
    
    for (FileUnit directory : directories) {
      /* get contents (files and directories) of the given directory */
      UserIO.print(UserIO.out.INFO,"Scanning directory "+directory.nativeName+"...",false);
      long size = FileIO.browseDirectory(directory,directory,dirContents);
      dirSize += size;
      /* count the number of items of the last source directory */
      dirCount = dirContents.size() - dirCount;
      UserIO.print(UserIO.out.INFO,dirCount + " items ("+FileIO.getSize(size)+")");
      dirCount = dirContents.size();
    }
    /* sort in lexicographic order of the name field to facilitate upcoming searches */
    Collections.sort(dirContents);
    
    UserIO.print(UserIO.out.INFO,"Total items: " + dirContents.size()+" ("+FileIO.getSize(dirSize)+")");
    return dirSize;
    
  }
  

  /**
   * Get directory contents.
   */
  public static long browseDirectory(
                                     final FileUnit rootDir,
                                     final FileUnit dir,
                                     final List<FileUnit> dirContents)
  {
    /* files and subdirectories of the specified directory */
    final List<File> files = new ArrayList<File>();
    final List<File> dirs = new ArrayList<File>();
    long size = 0;
    UserIO.print(UserIO.out.VERBOSE,"Listed directory "+dir.nativeName);
    /* prevent loops */
	if (dirContents.contains(dir)) {
		UserIO.reportWarning(UserIO.out.INFO,"Skipping directory " +dir.file.getPath()+ ": loop detected");
		return -1;
	}
			
    /* get directory contents */
    final FileIO.FileIOResult result = FileIO.listAllFiles(dir.file, files, dirs);
    if (!result.success) {
      UserIO.reportWarning(UserIO.out.INFO,"Unable to read contents from directory " +dir.file.getPath()+ ": " +result.errorMessage);
      return -1;
    }
    /* handle the directory */

    final FileUnit ud = new FileUnit(shadow2real(rootDir.file.getPath()), /* native pathname */
                                     shadow2real(dir.file.getPath().replace(rootDir.file.getPath(),"")), /* filePath */
                                     true);   
    if (ud.matchesFilter()) {
      ud.size = 0L;
      ud.time = dir.file.lastModified();
      dirContents.add(ud);
    }
    
    for (File f: files) {
      /* handle the files */
    	// normalizeUnicode(
      final FileUnit u = new FileUnit(shadow2real(rootDir.file.getPath()),
                                      shadow2real(f.getPath().replace(rootDir.file.getPath(),"")),
                                      false);       /* File object representing this file */
      /* apply filter */
      if (u.matchesFilter()) {
        u.size = f.length();
        size +=u.size;
        u.time = f.lastModified();
        dirContents.add(u);
        UserIO.print(UserIO.out.VERBOSE,"Listed "+u.nativeName + " ("+FileIO.getSize(u.size)+")");
      } else {
        UserIO.print(UserIO.out.VERBOSE,"Skipped file "+u.nativeName + " ("+FileIO.getSize(u.size)+")");
      }
    }
    /* recurse into subdirectories */
    for (File directoryFile : dirs) {
    	FileUnit directory = new FileUnit(shadow2real(rootDir.file.getPath()),shadow2real(directoryFile.getPath().replace(rootDir.file.getPath(),"")),true);
    	size += browseDirectory(rootDir,directory, dirContents);
    }
    return size;
  }
  
  
  /**
   * Get directory contents.
   */
  public static FileIOResult listAllFiles(final File dir,final List<File> files,final List<File> dirs)  {
    /* check if specified file/directory exists */
    if (!dir.exists()) return new FileIOResult(false,"Directory does not exist.");
    /* check if specified directory is a directory */
    if (!dir.isDirectory()) return new FileIOResult(false,"Specified directory is not a directory.");
    /* get list of files and directories in the specified directory */
    File listFiles[] = dir.listFiles();
    if (listFiles == null) {
    	return new FileIOResult(false,"Failed to list contents of directory.");
    } else {
    	for (File f : listFiles) {
    		/* skip symlinks */
    		if (isSymlink(f)) {
    			UserIO.reportWarning(UserIO.out.INFO,"Skipping symlink " +f.getPath());
    		} else {
    			if (f.isDirectory()) {
    				dirs.add(f);
    			} else {
    				files.add(f);
    			}    			
    		}
    	}
    	/* sort lists */
    	Collections.sort(files);
    	Collections.sort(dirs);
    	return new FileIOResult(true, null);
    }
  }
  
  
  /********************************
    * INNER CLASSES                *
    *******************************/
  
  /**
   * Inner class to represent the result of a file IO operation.
   */
  public static class FileIOResult {
    /** file IO operation is successful if true; unsuccessful otherwise */
    public boolean success = false;
    /** error message, if any */
    public String errorMessage = null;
    FileIOResult(final boolean success,final String errorMessage)
    {
      this.success = success;
      this.errorMessage = errorMessage;
    }
  }
  
  /**
   * Inner class to represent the result of a file CRC-32 computation.
   */
  public static class ChecksumResult {
    /** file CRC-32 computation is successful if true; unsuccessful otherwise */
    public boolean success = false;
    /** error message, if any */
    public String errorMessage = null;
    /** file CRC-32 checksum value */
    public String checksum;
    public ArrayList<SliceUnit> sliceList = new ArrayList<SliceUnit>();
    
    public ChecksumResult(final boolean success,
                          final String errorMessage,
                          final String checksum,
                          final ArrayList<SliceUnit> sliceList)
    {
      this.success = success;
      this.errorMessage = errorMessage;
      this.checksum = checksum;
      this.sliceList = sliceList;
    }
  }
  
  
}
