/**
 * TERMS AND CONDITIONS:
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.backup.zipsnap.ng.job.util;

import java.util.HashMap;
import java.util.Locale;
import java.util.Scanner;

import com.backup.zipsnap.ng.Main;
import com.backup.zipsnap.ng.gui.JobWrapper;
import com.backup.zipsnap.ng.gui.JobWrapper.msg;
import com.backup.zipsnap.ng.job.Config;
import com.backup.zipsnap.ng.job.Job;

/**
 * Simple class for basic user input/output.
 */
public class UserIO
{
	/* global variable */

	public static enum out { ERR,WARN,INFO,DEBUG,VERBOSE,ALL; }
	public static enum choice { YES_NO,YES_NO_ALWAYS_NEVER,OK; }
	public static ThreadLocal<HashMap<String,out>> warnings = new ThreadLocal<HashMap<String,out>>();

	/**
	 * Record and print a warning message
	 */
	public static void reportWarning(out type, String warningMessage) {
		if (! Config.initialized.get().booleanValue()) return;
		UserIO.print(type,"WARNING: " + warningMessage);
		Job.gui(JobWrapper.msg.MSG_WARNING, warningMessage);
		/* add the warning to the buffer */
		if (!warnings.get().containsKey(warningMessage)) warnings.get().put(warningMessage,type);
		/* ask the user for confirmation */
		if (!Main.settings.pauseOnWarning) return;
		UserIO.userCharPrompt(UserIO.choice.OK,warningMessage);
	}


	/**
	 * Print an error message and terminate program with default exit code.
	 */
	public static void reportError(String errorMessage)
	{
		if (! Config.initialized.get().booleanValue()) return;
		UserIO.print(UserIO.out.ERR,Main.lineSeparator+Main.lineSeparator+"ERROR: " + errorMessage);
		Job.gui(JobWrapper.msg.MSG_ERROR,errorMessage);
		/* close the log file before exiting */
		Job.exit(Main.settings.defaultExitCode);
	}


	/**
	 * Get custom exception message string for the given exception.
	 */
	public static String getExceptionMessage(final Exception e) {
		if (! Config.initialized.get().booleanValue()) return "";
		final StackTraceElement[] stackTrace = e.getStackTrace();
		final StringBuilder error = new StringBuilder();
		error.append("Exception information (" + e.getClass() +
				"):"+Main.lineSeparator+"\"" + e.getMessage() + "\"");
		for (int i = 0; i < stackTrace.length; i++) {
			error.append(Main.lineSeparator+"  at ");
			error.append(stackTrace[i]);
		}
		if (Main.settings.debug == 0) {
			if (Main.settings.logDebugAlways) Job.logger.get().log(error.toString());
			return e.getMessage();
		} else return error.toString();
	}

	/**
	 * Prompt user for a single-character input.
	 */
	public static char userCharPrompt(choice type,String question) {
		String options = null;
		String prompt = null;
		msg message = null;
		if (type == choice.YES_NO_ALWAYS_NEVER) {
			options = "YNAR";
			prompt = "(Y)es/(N)o/(A)lways/Neve(R): ";
			message = JobWrapper.msg.CONFIRM_OVERWRITE_DELETE;
		}
		else if (type == choice.YES_NO) {
			options = "YN";
			prompt = "(Y)es/(N)o: ";
			message = JobWrapper.msg.CONFIRM_SUMMARY;
		}
		else if (type == choice.OK) {
			options = "O";
			prompt = "(O)k: ";
			message = JobWrapper.msg.CONFIRM_OK;
		}
		UserIO.print(UserIO.out.INFO,question,false);
		try {
			@SuppressWarnings("resource")
			Scanner kb = new Scanner(System.in);
			while (true) {
				UserIO.print(UserIO.out.INFO,prompt,false);
				Job.gui(message,question);
				String choice = kb.nextLine().trim().toUpperCase(Locale.ENGLISH);
				if (choice.length() != 1) continue;
				/* convert to char */
				final char c = choice.charAt(0);
				if (options.indexOf(c) >= 0) {
					return c;
				}
			}
		} catch (Exception e) {
			UserIO.reportError("(FATAL) Cannot read from keyboard: " +UserIO.getExceptionMessage(e));
		}
		/* only on exception */
		return '\0';
	}


	/**
	 * Print a message to the given output
	 */
	public static void print(out type, String s) {
		print(type,s,true);
	}

	/**
	 * Print a message to the given output
	 */
	public static void print(out type, String s,boolean newline) {
		if (! Config.initialized.get().booleanValue()) return;
		if (newline) s = s+Main.lineSeparator;
		if (type == UserIO.out.ALL && Main.settings.debug < 3) return;
		if (type == UserIO.out.VERBOSE && Main.settings.debug < 2) return;
		if (Main.settings.logDebugAlways) {
			Job.logger.get().log(s);
			Job.gui(JobWrapper.msg.MSG_INFO,s);
		}
		if (type == UserIO.out.DEBUG && Main.settings.debug < 1) return;
		if (! Main.settings.logDebugAlways) {
			Job.logger.get().log(s);
			Job.gui(JobWrapper.msg.MSG_INFO,s);
		}
		if (!Job.isInGUI) {
			/* print to the standard output/error */
			if (type == UserIO.out.ERR) {
				System.err.print(s + "");
				System.err.flush();
			} else {
				System.out.print(s + "");
				System.out.flush();
			}
		}
	}

	/**
	 * Print the output justified (center)
	 */
	public static String centerJustify(final Object o,final int width) {
		final String s = o + "";
		final int len = s.length();
		final int totalSpace = width - len;
		if (totalSpace <= 0) return s;
		final StringBuilder t = new StringBuilder();
		final int leadingSpace = totalSpace / 2;
		for (int i = 0; i < leadingSpace; i++) t.append(' ');
		t.append(s);
		for (int i = 0; i < totalSpace - leadingSpace; i++) t.append(' ');
		return t.toString();
	}

}
